#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2021 Snowflake Computing Inc. All right reserved.
#

import base64
import hashlib
from datetime import datetime, timedelta
from logging import getLogger
from typing import Optional

import jwt
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    PublicFormat,
    load_der_private_key,
)

from .auth_by_plugin import AuthByPlugin
from .errorcode import ER_INVALID_PRIVATE_KEY
from .errors import ProgrammingError
from .network import KEY_PAIR_AUTHENTICATOR

logger = getLogger(__name__)


class AuthByKeyPair(AuthByPlugin):
    """Key pair based authentication."""

    ALGORITHM = "RS256"
    ISSUER = "iss"
    SUBJECT = "sub"
    EXPIRE_TIME = "exp"
    ISSUE_TIME = "iat"
    LIFETIME = 120

    def __init__(self, private_key, lifetime_in_seconds: int = LIFETIME):
        """Inits AuthByKeyPair class with private key.

        Args:
            private_key: a byte array of der formats of private key
            lifetime_in_seconds: number of seconds the JWT token will be valid
        """
        self._private_key = private_key
        self._jwt_token = ""
        self._jwt_token_exp = 0
        self._lifetime = timedelta(seconds=lifetime_in_seconds)

    def authenticate(
        self,
        authenticator: str,
        service_name: Optional[str],
        account: str,
        user: str,
        password: Optional[str],
    ) -> str:
        if ".global" in account:
            account = account.partition("-")[0]
        else:
            account = account.partition(".")[0]
        account = account.upper()
        user = user.upper()

        now = datetime.utcnow()

        try:
            private_key = load_der_private_key(
                data=self._private_key, password=None, backend=default_backend()
            )
        except Exception as e:
            raise ProgrammingError(
                msg="Failed to load private key: {}\nPlease provide a valid unencrypted rsa private "
                "key in DER format as bytes object".format(str(e)),
                errno=ER_INVALID_PRIVATE_KEY,
            )

        if not isinstance(private_key, RSAPrivateKey):
            raise ProgrammingError(
                msg="Private key type ({}) not supported.\nPlease provide a valid rsa private "
                "key in DER format as bytes object".format(
                    private_key.__class__.__name__
                ),
                errno=ER_INVALID_PRIVATE_KEY,
            )

        public_key_fp = self.calculate_public_key_fingerprint(private_key)

        self._jwt_token_exp = now + self._lifetime
        payload = {
            self.ISSUER: "{}.{}.{}".format(account, user, public_key_fp),
            self.SUBJECT: "{}.{}".format(account, user),
            self.ISSUE_TIME: now,
            self.EXPIRE_TIME: self._jwt_token_exp,
        }

        _jwt_token = jwt.encode(payload, private_key, algorithm=self.ALGORITHM)

        # jwt.encode() returns bytes in pyjwt 1.x and a string
        # in pyjwt 2.x
        if isinstance(_jwt_token, bytes):
            self._jwt_token = _jwt_token.decode("utf-8")
        else:
            self._jwt_token = _jwt_token

        return self._jwt_token

    @staticmethod
    def calculate_public_key_fingerprint(private_key):
        # get public key bytes
        public_key_der = private_key.public_key().public_bytes(
            Encoding.DER, PublicFormat.SubjectPublicKeyInfo
        )

        # take sha256 on raw bytes and then do base64 encode
        sha256hash = hashlib.sha256()
        sha256hash.update(public_key_der)

        public_key_fp = "SHA256:" + base64.b64encode(sha256hash.digest()).decode(
            "utf-8"
        )
        logger.debug("Public key fingerprint is %s", public_key_fp)

        return public_key_fp

    def update_body(self, body):
        body["data"]["AUTHENTICATOR"] = KEY_PAIR_AUTHENTICATOR
        body["data"]["TOKEN"] = self._jwt_token

    def assertion_content(self):
        return self._jwt_token



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2021 Snowflake Computing Inc. All right reserved.
#
# Feature flags

feature_use_pyopenssl = True  # use pyopenssl API or openssl command



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2021 Snowflake Computing Inc. All right reserved.
#

import json
import logging
import socket
import time
import webbrowser

from .auth import Auth
from .auth_by_plugin import AuthByPlugin
from .compat import parse_qs, urlparse, urlsplit
from .constants import (
    HTTP_HEADER_ACCEPT,
    HTTP_HEADER_CONTENT_TYPE,
    HTTP_HEADER_SERVICE_NAME,
    HTTP_HEADER_USER_AGENT,
)
from .errorcode import (
    ER_IDP_CONNECTION_ERROR,
    ER_NO_HOSTNAME_FOUND,
    ER_UNABLE_TO_OPEN_BROWSER,
)
from .errors import OperationalError
from .network import (
    CONTENT_TYPE_APPLICATION_JSON,
    EXTERNAL_BROWSER_AUTHENTICATOR,
    PYTHON_CONNECTOR_USER_AGENT,
)

logger = logging.getLogger(__name__)

BUF_SIZE = 16384


# global state of web server that receives the SAML assertion from
# Snowflake server


class AuthByWebBrowser(AuthByPlugin):
    """Authenticates user by web browser. Only used for SAML based authentication."""

    def __init__(
        self,
        rest,
        application,
        webbrowser_pkg=None,
        socket_pkg=None,
        protocol=None,
        host=None,
        port=None,
    ):
        self._rest = rest
        self._token = None
        self._consent_cache_id_token = True
        self._application = application
        self._proof_key = None
        self._webbrowser = webbrowser if webbrowser_pkg is None else webbrowser_pkg
        self._socket = socket.socket if socket_pkg is None else socket_pkg
        self._protocol = protocol
        self._host = host
        self._port = port
        self._origin = None

    @property
    def consent_cache_id_token(self):
        return self._consent_cache_id_token

    @property
    def assertion_content(self):
        """Returns the token."""
        return self._token

    def update_body(self, body):
        """Used by Auth to update the request that gets sent to /v1/login-request.

        Args:
            body: existing request dictionary
        """
        body["data"]["AUTHENTICATOR"] = EXTERNAL_BROWSER_AUTHENTICATOR
        body["data"]["TOKEN"] = self._token
        body["data"]["PROOF_KEY"] = self._proof_key

    def authenticate(self, authenticator, service_name, account, user, password):
        """Web Browser based Authentication."""
        logger.debug("authenticating by Web Browser")

        # ignore password. user is still needed by GS to verify
        # the assertion.
        _ = password  # noqa: F841

        socket_connection = self._socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            try:
                socket_connection.bind(("localhost", 0))
            except socket.gaierror as ex:
                if ex.args[0] == socket.EAI_NONAME:
                    raise OperationalError(
                        msg="localhost is not found. Ensure /etc/hosts has "
                        "localhost entry.",
                        errno=ER_NO_HOSTNAME_FOUND,
                    )
                else:
                    raise ex
            socket_connection.listen(0)  # no backlog
            callback_port = socket_connection.getsockname()[1]

            print(
                "Initiating login request with your identity provider. A "
                "browser window should have opened for you to complete the "
                "login. If you can't see it, check existing browser windows, "
                "or your OS settings. Press CTRL+C to abort and try again..."
            )

            logger.debug("step 1: query GS to obtain SSO url")
            sso_url = self._get_sso_url(
                authenticator, service_name, account, callback_port, user
            )

            logger.debug("step 2: open a browser")
            if not self._webbrowser.open_new(sso_url):
                logger.error(
                    "Unable to open a browser in this environment.", exc_info=True
                )
                print(
                    "We were unable to open a browser window for you, "
                    f"please open the following url manually: {sso_url}"
                )
                self.handle_failure(
                    {
                        "code": ER_UNABLE_TO_OPEN_BROWSER,
                        "message": "Unable to open a browser in this environment.",
                    }
                )
                return  # required for test case

            logger.debug("step 3: accept SAML token")
            self._receive_saml_token(socket_connection)
        finally:
            socket_connection.close()

    def _receive_saml_token(self, socket_connection):
        """Receives SAML token from web browser."""
        while True:
            socket_client, _ = socket_connection.accept()
            try:
                # Receive the data in small chunks and retransmit it
                data = socket_client.recv(BUF_SIZE).decode("utf-8").split("\r\n")

                if not self._process_options(data, socket_client):
                    self._process_receive_saml_token(data, socket_client)
                    break
            finally:
                socket_client.shutdown(socket.SHUT_RDWR)
                socket_client.close()

    def _process_options(self, data, socket_client):
        """Allows JS Ajax access to this endpoint."""
        for line in data:
            if line.startswith("OPTIONS "):
                break
        else:
            return False

        self._get_user_agent(data)
        requested_headers, requested_origin = self._check_post_requested(data)
        if not requested_headers:
            return False

        if not self._validate_origin(requested_origin):
            # validate Origin and fail if not match with the server.
            return False

        self._origin = requested_origin
        content = [
            "HTTP/1.1 200 OK",
            "Date: {}".format(
                time.strftime("%a, %d %b %Y %H:%M:%S GMT", time.gmtime())
            ),
            "Access-Control-Allow-Methods: POST, GET",
            "Access-Control-Allow-Headers: {}".format(requested_headers),
            "Access-Control-Max-Age: 86400",
            "Access-Control-Allow-Origin: {}".format(self._origin),
            "",
            "",
        ]
        socket_client.sendall("\r\n".join(content).encode("utf-8"))
        return True

    def _validate_origin(self, requested_origin):
        ret = urlsplit(requested_origin)
        netloc = ret.netloc.split(":")
        host_got = netloc[0]
        port_got = (
            netloc[1] if len(netloc) > 1 else (443 if self._protocol == "https" else 80)
        )

        return (
            ret.scheme == self._protocol
            and host_got == self._host
            and port_got == self._port
        )

    def _process_receive_saml_token(self, data, socket_client):
        if not self._process_get(data) and not self._process_post(data):
            return  # error

        content = [
            "HTTP/1.1 200 OK",
            "Content-Type: text/html",
        ]
        if self._origin:
            data = {"consent": self._consent_cache_id_token}
            msg = json.dumps(data)
            content.append("Access-Control-Allow-Origin: {}".format(self._origin))
            content.append("Vary: Accept-Encoding, Origin")
        else:
            msg = """
<!DOCTYPE html><html><head><meta charset="UTF-8"/>
<title>SAML Response for Snowflake</title></head>
<body>
Your identity was confirmed and propagated to Snowflake {}.
You can close this window now and go back where you started from.
</body></html>""".format(
                self._application
            )
        content.append("Content-Length: {}".format(len(msg)))
        content.append("")
        content.append(msg)

        socket_client.sendall("\r\n".join(content).encode("utf-8"))

    def _check_post_requested(self, data):
        request_line = None
        header_line = None
        origin_line = None
        for line in data:
            if line.startswith("Access-Control-Request-Method:"):
                request_line = line
            elif line.startswith("Access-Control-Request-Headers:"):
                header_line = line
            elif line.startswith("Origin:"):
                origin_line = line

        if (
            not request_line
            or not header_line
            or not origin_line
            or request_line.split(":")[1].strip() != "POST"
        ):
            return None, None

        return (
            header_line.split(":")[1].strip(),
            ":".join(origin_line.split(":")[1:]).strip(),
        )

    def _process_get(self, data):
        for line in data:
            if line.startswith("GET "):
                target_line = line
                break
        else:
            return False

        self._get_user_agent(data)
        _, url, _ = target_line.split()
        self._token = parse_qs(urlparse(url).query)["token"][0]
        return True

    def _process_post(self, data):
        for line in data:
            if line.startswith("POST "):
                break
        else:
            self.handle_failure(
                {
                    "code": ER_IDP_CONNECTION_ERROR,
                    "message": "Invalid HTTP request from web browser. Idp "
                    "authentication could have failed.",
                }
            )
            return False

        self._get_user_agent(data)
        try:
            # parse the response as JSON
            payload = json.loads(data[-1])
            self._token = payload.get("token")
            self._consent_cache_id_token = payload.get("consent", True)
        except Exception:
            # key=value form.
            self._token = parse_qs(data[-1])["token"][0]
        return True

    def _get_user_agent(self, data):
        for line in data:
            if line.lower().startswith("user-agent"):
                logger.debug(line)
                break
        else:
            logger.debug("No User-Agent")

    def _get_sso_url(self, authenticator, service_name, account, callback_port, user):
        """Gets SSO URL from Snowflake."""
        headers = {
            HTTP_HEADER_CONTENT_TYPE: CONTENT_TYPE_APPLICATION_JSON,
            HTTP_HEADER_ACCEPT: CONTENT_TYPE_APPLICATION_JSON,
            HTTP_HEADER_USER_AGENT: PYTHON_CONNECTOR_USER_AGENT,
        }
        if service_name:
            headers[HTTP_HEADER_SERVICE_NAME] = service_name

        url = "/session/authenticator-request"
        body = Auth.base_auth_data(
            user,
            account,
            self._rest._connection.application,
            self._rest._connection._internal_application_name,
            self._rest._connection._internal_application_version,
            self._rest._connection._ocsp_mode(),
            self._rest._connection._login_timeout,
            self._rest._connection._network_timeout,
        )

        body["data"]["AUTHENTICATOR"] = authenticator
        body["data"]["BROWSER_MODE_REDIRECT_PORT"] = str(callback_port)
        logger.debug(
            "account=%s, authenticator=%s, user=%s", account, authenticator, user
        )
        ret = self._rest._post_request(
            url,
            headers,
            json.dumps(body),
            timeout=self._rest._connection.login_timeout,
            socket_timeout=self._rest._connection.login_timeout,
        )
        if not ret["success"]:
            self.handle_failure(ret)
        data = ret["data"]
        sso_url = data["ssoUrl"]
        self._proof_key = data["proofKey"]
        return sso_url



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2021 Snowflake Computing Inc. All right reserved.
#

import binascii
import glob
import mimetypes
import os
import shutil
import sys
import tempfile
import threading
from concurrent.futures.thread import ThreadPoolExecutor
from logging import getLogger
from queue import Queue
from time import sleep, time
from typing import IO, TYPE_CHECKING, Any, Dict, List, Optional, Type, Union

from boto3.session import Session

from .compat import GET_CWD, IS_WINDOWS, dataclass, field
from .constants import ResultStatus, megabyte
from .converter_snowsql import SnowflakeConverterSnowSQL
from .errorcode import (
    ER_COMPRESSION_NOT_SUPPORTED,
    ER_FAILED_TO_DOWNLOAD_FROM_STAGE,
    ER_FAILED_TO_UPLOAD_TO_STAGE,
    ER_FILE_NOT_EXISTS,
    ER_INTERNAL_NOT_MATCH_ENCRYPT_MATERIAL,
    ER_INVALID_STAGE_FS,
    ER_INVALID_STAGE_LOCATION,
    ER_LOCAL_PATH_NOT_DIRECTORY,
)
from .errors import (
    DatabaseError,
    Error,
    InternalError,
    OperationalError,
    ProgrammingError,
)
from .file_compression_type import CompressionTypes, lookup_by_mime_sub_type
from .file_util import SnowflakeFileUtil
from .gcs_util_sdk import SnowflakeGCSUtil
from .local_util_sdk import SnowflakeLocalUtil
from .remote_storage_util_sdk import (
    SnowflakeFileEncryptionMaterial,
    SnowflakeRemoteStorageUtil,
)
from .s3_util_sdk import SnowflakeS3Util

if TYPE_CHECKING:  # pragma: no cover
    from azure.storage.blob import BlobServiceClient

    from .cursor import SnowflakeCursor
    from .file_compression_type import CompressionType

S3_FS = "S3"
AZURE_FS = "AZURE"
GCS_FS = "GCS"
LOCAL_FS = "LOCAL_FS"
CMD_TYPE_UPLOAD = "UPLOAD"
CMD_TYPE_DOWNLOAD = "DOWNLOAD"
FILE_PROTOCOL = "file://"

INJECT_WAIT_IN_PUT = 0

logger = getLogger(__name__)


def result_text_column_desc(name):
    return {
        "name": name,
        "type": "text",
        "length": 16777216,
        "precision": None,
        "scale": None,
        "nullable": False,
    }


def result_fixed_column_desc(name):
    return {
        "name": name,
        "type": "fixed",
        "length": 5,
        "precision": 0,
        "scale": 0,
        "nullable": False,
    }


@dataclass
class SFResourceMeta:
    storage_util: Optional[
        Type[Union["SnowflakeRemoteStorageUtil", "SnowflakeLocalUtil"]]
    ] = None
    stage_info: Optional[Dict[str, Any]] = field(
        default_factory=dict
    )  # TODO could be strongly defined if need be
    use_accelerate_endpoint: bool = False
    cloud_client: Union["Session.resource", "BlobServiceClient", str, None] = None
    use_s3_regional_url: bool = False


@dataclass
class SnowflakeFileMeta:
    """Class to keep track of information necessary for file operations."""

    name: str
    src_file_name: str
    stage_location_type: str
    result_status: Optional["ResultStatus"] = None

    # client_meta holds cloud client object and the metadata needed to create such client
    client_meta: "SFResourceMeta" = SFResourceMeta(cloud_client=None)
    self: Optional["SnowflakeFileTransferAgent"] = None
    put_callback: Optional[Type["SnowflakeProgressPercentage"]] = None
    put_azure_callback: Optional[Type["SnowflakeProgressPercentage"]] = None
    put_callback_output_stream: Optional[IO[str]] = None
    get_callback: Optional[Type["SnowflakeProgressPercentage"]] = None
    get_azure_callback: Optional[Type["SnowflakeProgressPercentage"]] = None
    get_callback_output_stream: Optional[IO[str]] = None
    show_progress_bar: bool = False
    multipart_threshold: int = 67108864  # Historical value
    parallel: int = 1
    presigned_url: Optional[str] = None
    overwrite: bool = False
    tmp_dir: Optional[str] = None
    sha256_digest: Optional[str] = None
    upload_size: Optional[int] = None
    real_src_file_name: Optional[str] = None
    error_details: Optional[str] = None
    last_max_concurrency: int = -1
    last_error: Optional[Exception] = None
    no_sleeping_time: bool = False
    gcs_file_header_digest: Optional[str] = None
    gcs_file_header_content_length: Optional[int] = None
    gcs_file_header_encryption_metadata: Optional[Dict[str, Any]] = None

    encryption_material: Optional["SnowflakeFileEncryptionMaterial"] = None
    tmp_dir: Optional[str] = None
    # Specific to Uploads only
    src_file_size: int = 0
    src_compression_type: Optional["CompressionType"] = None
    dst_compression_type: "CompressionType" = None
    require_compress: bool = False
    dst_file_name: Optional[str] = None
    dst_file_size: int = -1
    src_stream: Optional[IO[bytes]] = None
    real_src_stream: Optional[IO[bytes]] = None
    # Specific to Downloads only
    local_location: Optional[str] = None


def _update_progress(
    file_name: str,
    start_time: float,
    total_size: float,
    progress: Union[float, int],
    output_stream: Optional[IO] = sys.stdout,
    show_progress_bar: Optional[bool] = True,
) -> bool:
    bar_length = 10  # Modify this to change the length of the progress bar
    total_size /= megabyte
    status = ""
    elapsed_time = time() - start_time
    throughput = (total_size / elapsed_time) if elapsed_time != 0.0 else 0.0
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
        status = "error: progress var must be float\r\n"
    if progress < 0:
        progress = 0
        status = "Halt...\r\n"
    if progress >= 1:
        progress = 1
        status = f"Done ({elapsed_time:.3f}s, {throughput:.2f}MB/s).\r\n"
    if not status and show_progress_bar:
        status = f"({elapsed_time:.3f}s, {throughput:.2f}MB/s)"
    if status:
        block = int(round(bar_length * progress))
        text = (
            f"\r{file_name}({total_size:.2f}MB): "
            f"[{'#' * block + '-' * (bar_length - block)}] "
            f"{progress * 100.0:.2f}% {status}"
        )
        output_stream.write(text)
        output_stream.flush()
    logger.debug(
        f"filename: {file_name}, start_time: {start_time}, total_size: {total_size}, "
        f"progress: {progress}, show_progress_bar: {show_progress_bar}"
    )
    return progress == 1.0


def percent(seen_so_far: int, size: float) -> float:
    return 1.0 if seen_so_far >= size or size <= 0 else float(seen_so_far / size)


class SnowflakeProgressPercentage:
    """Built-in Progress bar for PUT commands."""

    def __init__(
        self,
        filename: str,
        filesize: Union[int, float],
        output_stream: Optional[IO] = sys.stdout,
        show_progress_bar: Optional[bool] = True,
    ):
        last_pound_char = filename.rfind("#")
        if last_pound_char < 0:
            last_pound_char = len(filename)
        self._filename = os.path.basename(filename[0:last_pound_char])
        self._output_stream = output_stream
        self._show_progress_bar = show_progress_bar
        self._size = float(filesize)
        self._seen_so_far = 0
        self._done = False
        self._start_time = time()
        self._lock = threading.Lock()

    def __call__(self, bytes_amount: int):
        raise NotImplementedError


class SnowflakeS3ProgressPercentage(SnowflakeProgressPercentage):
    def __init__(
        self,
        filename: str,
        filesize: Union[int, float],
        output_stream: Optional[IO] = sys.stdout,
        show_progress_bar: Optional[bool] = True,
    ):
        super(SnowflakeS3ProgressPercentage, self).__init__(
            filename,
            filesize,
            output_stream=output_stream,
            show_progress_bar=show_progress_bar,
        )

    def __call__(self, bytes_amount: int):
        with self._lock:
            if self._output_stream:
                self._seen_so_far += bytes_amount
                percentage = percent(self._seen_so_far, self._size)
                if not self._done:
                    self._done = _update_progress(
                        self._filename,
                        self._start_time,
                        self._size,
                        percentage,
                        output_stream=self._output_stream,
                        show_progress_bar=self._show_progress_bar,
                    )


class SnowflakeAzureProgressPercentage(SnowflakeProgressPercentage):
    def __init__(
        self,
        filename: str,
        filesize: Union[int, float],
        output_stream: Optional[IO] = sys.stdout,
        show_progress_bar: Optional[bool] = True,
    ):
        super(SnowflakeAzureProgressPercentage, self).__init__(
            filename,
            filesize,
            output_stream=output_stream,
            show_progress_bar=show_progress_bar,
        )

    def __call__(self, current: int):
        with self._lock:
            if self._output_stream:
                self._seen_so_far = current
                percentage = percent(self._seen_so_far, self._size)
                if not self._done:
                    self._done = _update_progress(
                        self._filename,
                        self._start_time,
                        self._size,
                        percentage,
                        output_stream=self._output_stream,
                        show_progress_bar=self._show_progress_bar,
                    )


class SnowflakeFileTransferAgent(object):
    """Snowflake File Transfer Agent provides cloud provider independent implementation for putting/getting files."""

    def __init__(
        self,
        cursor: "SnowflakeCursor",
        command: str,
        ret: Dict[str, Any],
        put_callback: Optional[Type["SnowflakeProgressPercentage"]] = None,
        put_azure_callback: Optional[Type["SnowflakeProgressPercentage"]] = None,
        put_callback_output_stream: IO[str] = sys.stdout,
        get_callback: Optional[Type["SnowflakeProgressPercentage"]] = None,
        get_azure_callback: Optional[Type["SnowflakeProgressPercentage"]] = None,
        get_callback_output_stream: IO[str] = sys.stdout,
        show_progress_bar: bool = True,
        raise_put_get_error: bool = True,
        force_put_overwrite: bool = True,
        multipart_threshold: Optional[int] = None,
        source_from_stream: Optional[IO[bytes]] = None,
        use_s3_regional_url: bool = False,
    ):
        self._cursor = cursor
        self._command = command
        self._ret = ret
        self._put_callback = put_callback
        self._put_azure_callback = (
            put_azure_callback if put_azure_callback else put_callback
        )
        self._put_callback_output_stream = put_callback_output_stream
        self._get_callback = get_callback
        self._get_azure_callback = (
            get_azure_callback if get_azure_callback else get_callback
        )
        self._get_callback_output_stream = get_callback_output_stream
        self._use_accelerate_endpoint = False
        self._raise_put_get_error = raise_put_get_error
        self._show_progress_bar = show_progress_bar
        self._force_put_overwrite = force_put_overwrite
        self._source_from_stream = source_from_stream
        # The list of self-sufficient file metas that are sent to
        # remote storage clients to get operated on.
        self._file_metadata: List["SnowflakeFileMeta"] = []
        self._results: List["SnowflakeFileMeta"] = []
        if multipart_threshold is not None:
            self._multipart_threshold = multipart_threshold
        else:
            # Historical value
            self._multipart_threshold = 67108864
        self._use_s3_regional_url = use_s3_regional_url

    def execute(self):
        self._parse_command()

        self._init_file_metadata()

        if self._command_type == CMD_TYPE_UPLOAD:
            self._process_file_compression_type()

        self._transfer_accelerate_config()

        if self._command_type == CMD_TYPE_DOWNLOAD:
            if not os.path.isdir(self._local_location):
                os.makedirs(self._local_location)

        if self._stage_location_type == LOCAL_FS:
            if not os.path.isdir(self._stage_info["location"]):
                os.makedirs(self._stage_info["location"])

        self._update_file_metas_with_presigned_url()

        small_file_metas = []
        large_file_metas = []
        for m in self._file_metadata:
            m.overwrite = self._overwrite
            m.self = self
            if self._stage_location_type != LOCAL_FS:
                m.put_callback = self._put_callback
                m.put_azure_callback = self._put_azure_callback
                m.put_callback_output_stream = self._put_callback_output_stream
                m.get_callback = self._get_callback
                m.get_azure_callback = self._get_azure_callback
                m.get_callback_output_stream = self._get_callback_output_stream
                m.show_progress_bar = self._show_progress_bar

                # multichunk uploader threshold
                size_threshold = self._multipart_threshold
                m.multipart_threshold = size_threshold
                if m.src_file_size > size_threshold:
                    m.parallel = self._parallel
                    large_file_metas.append(m)
                else:
                    m.parallel = 1
                    small_file_metas.append(m)
            else:
                m.parallel = 1
                small_file_metas.append(m)

        logger.debug(f"parallel=[{self._parallel}]")
        if self._command_type == CMD_TYPE_UPLOAD:
            self.upload(large_file_metas, small_file_metas)
        else:
            self.download(large_file_metas, small_file_metas)

        # turn enum to string, in order to have backward compatible interface
        for result in self._results:
            result.result_status = result.result_status.value

    def upload(
        self,
        large_file_metas: List["SnowflakeFileMeta"],
        small_file_metas: List["SnowflakeFileMeta"],
    ):
        storage_util = SnowflakeFileTransferAgent.get_storage_client(
            self._stage_location_type
        )
        client_meta = SFResourceMeta(
            storage_util=storage_util,
            stage_info=self._stage_info,
            use_accelerate_endpoint=self._use_accelerate_endpoint,
            use_s3_regional_url=self._use_s3_regional_url,
        )

        for meta in small_file_metas:
            meta.client_meta = client_meta
        for meta in large_file_metas:
            meta.client_meta = client_meta

        if len(small_file_metas) > 0:
            self._upload_files_in_parallel(small_file_metas)
        if len(large_file_metas) > 0:
            self._upload_files_in_sequential(large_file_metas)

    def _transfer_accelerate_config(self):
        if self._stage_location_type == S3_FS:
            client = SnowflakeRemoteStorageUtil.create_client(
                self._stage_info,
                use_accelerate_endpoint=False,
                use_s3_regional_url=self._use_s3_regional_url,
                s3_connection_pool_size=self._cursor._connection._s3_connection_pool_size,
            )
            self._use_accelerate_endpoint = SnowflakeS3Util.transfer_accelerate_config(
                client, self._stage_info
            )

    def _upload_files_in_parallel(self, file_metas: List["SnowflakeFileMeta"]) -> None:
        """Uploads files in parallel.

        Args:
            file_metas: List of metadata for files to be uploaded.
        """
        qtask: Queue["SnowflakeFileMeta"] = Queue()
        for meta in file_metas:
            qtask.put(meta)
        retry_round = 0
        presigned_url_handled = False

        while not qtask.empty():
            len_file_metas = qtask.qsize()
            thread_number = min(len_file_metas, self._parallel)
            logger.debug(
                f"start {retry_round} round for {qtask.qsize()} items with {thread_number} threads"
            )
            pool = ThreadPoolExecutor(thread_number)
            triggers = Queue()
            for _ in range(thread_number):
                pool.submit(
                    SnowflakeFileTransferAgent._upload_files_in_queue_thread,
                    qtask,
                    self._results,
                    triggers,
                )
            pool.shutdown()

            # update presigned url before starting next round
            # every time we see renew_token triggers, do one renew only
            renew_token_handled = False
            while not triggers.empty():
                result_meta = triggers.get()
                if (
                    not presigned_url_handled
                    and result_meta.result_status == ResultStatus.RENEW_PRESIGNED_URL
                ):
                    self._update_file_metas_with_presigned_url()
                    presigned_url_handled = True
                if (
                    not renew_token_handled
                    and result_meta.result_status == ResultStatus.RENEW_TOKEN
                ):
                    logger.debug("renewing expired token")
                    ret = self._cursor._execute_helper(
                        self._command
                    )  # rerun the command to get the credential
                    self._stage_info = ret["data"]["stageInfo"]

                    while qtask.not_empty:
                        meta = qtask.get()
                        meta.client_meta.stage_info = self._stage_info
                    renew_token_handled = True
            retry_round += 1

    @staticmethod
    def _upload_files_in_queue_thread(
        qtask: "Queue[SnowflakeFileMeta]",
        results: "List[SnowflakeFileMeta]",
        triggers: "Queue[SnowflakeFileMeta]",
    ):
        """Worker function for a single thread to do parallel uploading.

        qtask: the queue contains tasks to be processed.
        qnext: the queue for tasks to retry in next round.
        results: the results of finished tasks.
        triggers: when there is results trigger extra actions after this round of threads are done,
             e.g. RENEW_PRESIGNED_URL,
        """
        logger.debug(
            f"Enter upload worker thread. tid={threading.current_thread().ident}"
        )
        thread_client = None
        while not qtask.empty():
            # triggers is not empty means a result trigger all threads to quit
            if not triggers.empty():
                break

            meta = qtask.get()
            # initialize resource if it is not initialized yet
            cln_meta = meta.client_meta
            if thread_client is None:
                thread_client = cln_meta.storage_util.create_client(
                    cln_meta.stage_info,
                    cln_meta.use_accelerate_endpoint,
                    cln_meta.use_s3_regional_url,
                    s3_connection_pool_size=meta.self._cursor._connection._s3_connection_pool_size,
                )
            cln_meta.cloud_client = thread_client
            result_meta = SnowflakeFileTransferAgent.upload_one_file(meta)
            if result_meta.result_status == ResultStatus.RENEW_TOKEN:
                # need to retry this upload. the meta will be added back once renew is done
                triggers.put(result_meta)
                qtask.put(meta)
            elif result_meta.result_status == ResultStatus.RENEW_PRESIGNED_URL:
                # now stop this round - by adding one item to triggers
                triggers.put(result_meta)
                qtask.put(meta)
            else:
                results.append(result_meta)

    def _upload_files_in_sequential(self, file_metas: List["SnowflakeFileMeta"]):
        """Uploads files in sequential. Retry if the access token expires.

        Args:
            file_metas: List of metadata for files to be uploaded.
        """
        idx = 0
        len_file_metas = len(file_metas)

        cln_meta = file_metas[0].client_meta
        client = cln_meta.storage_util.create_client(
            cln_meta.stage_info,
            use_accelerate_endpoint=cln_meta.use_accelerate_endpoint,
            use_s3_regional_url=cln_meta.use_s3_regional_url,
            s3_connection_pool_size=self._cursor._connection._s3_connection_pool_size,
        )
        for meta in file_metas:
            meta.client_meta.cloud_client = client
        while idx < len_file_metas:
            logger.debug(f"uploading files idx: {idx+1}/{len_file_metas}")
            result = SnowflakeFileTransferAgent.upload_one_file(file_metas[idx])
            if result.result_status == ResultStatus.RENEW_TOKEN:
                client = self.renew_expired_client()
                for idx0 in range(idx, len_file_metas):
                    file_metas[idx0].client_meta.cloud_client = client
                continue
            elif result.result_status == ResultStatus.RENEW_PRESIGNED_URL:
                self._update_file_metas_with_presigned_url()
                continue
            self._results.append(result)
            idx += 1
            if INJECT_WAIT_IN_PUT > 0:
                logger.debug(f"LONGEVITY TEST: waiting for {INJECT_WAIT_IN_PUT}")
                sleep(INJECT_WAIT_IN_PUT)

    @staticmethod
    def get_storage_client(stage_location_type):
        if stage_location_type == LOCAL_FS:
            return SnowflakeLocalUtil
        elif stage_location_type in [S3_FS, AZURE_FS, GCS_FS]:
            return SnowflakeRemoteStorageUtil
        else:
            return None

    @staticmethod
    def upload_one_file(meta: "SnowflakeFileMeta") -> "SnowflakeFileMeta":
        """Uploads one file.

        Args:
            meta: Metadata for file to be uploaded.

        Returns:
            Metadata of uploaded file.
        """
        logger.debug(f"uploading file={meta.src_file_name}")
        meta.real_src_file_name = meta.src_file_name
        tmp_dir = tempfile.mkdtemp()
        meta.tmp_dir = tmp_dir
        try:
            if meta.require_compress:
                logger.debug(f"compressing file={meta.src_file_name}")
                if meta.src_stream is None:
                    (
                        meta.real_src_file_name,
                        upload_size,
                    ) = SnowflakeFileUtil.compress_file_with_gzip(
                        meta.src_file_name, tmp_dir
                    )
                else:
                    (
                        meta.real_src_stream,
                        upload_size,
                    ) = SnowflakeFileUtil.compress_with_gzip_from_stream(
                        meta.src_stream
                    )

            logger.debug(f"getting digest file={meta.real_src_file_name}")
            if meta.src_stream is None:
                (
                    meta.sha256_digest,
                    meta.upload_size,
                ) = SnowflakeFileUtil.get_digest_and_size_for_file(
                    meta.real_src_file_name
                )
            else:
                (
                    meta.sha256_digest,
                    meta.upload_size,
                ) = SnowflakeFileUtil.get_digest_and_size_for_stream(
                    meta.real_src_stream or meta.src_stream
                )
            logger.debug("really uploading data")
            storage_client = SnowflakeFileTransferAgent.get_storage_client(
                meta.stage_location_type
            )
            storage_client.upload_one_file_with_retry(meta)
            logger.debug(
                f"done: status={meta.result_status}, file={meta.src_file_name}, real file={meta.real_src_file_name}"
            )
        except Exception as e:
            logger.exception(
                f"Failed to upload a file: file={meta.src_file_name}, real file={meta.real_src_file_name}"
            )
            meta.dst_file_size = 0
            if meta.result_status is None:
                meta.result_status = ResultStatus.ERROR
            meta.error_details = str(e)
            meta.error_details += (
                f", file={meta.src_file_name}, real file={meta.real_src_file_name}"
            )
        finally:
            logger.debug(f"cleaning up tmp dir: {tmp_dir}")
            shutil.rmtree(tmp_dir)
            if meta.src_stream is not None:
                meta.src_stream.seek(0)
            if meta.real_src_stream is not None:
                meta.real_src_stream.close()
        return meta

    def download(
        self,
        large_file_metas: List["SnowflakeFileMeta"],
        small_file_metas: List["SnowflakeFileMeta"],
    ):
        storage_client = SnowflakeFileTransferAgent.get_storage_client(
            self._stage_location_type
        )
        client = storage_client.create_client(
            self._stage_info,
            use_accelerate_endpoint=self._use_accelerate_endpoint,
            use_s3_regional_url=self._use_s3_regional_url,
            s3_connection_pool_size=self._cursor._connection._s3_connection_pool_size,
        )
        for meta in small_file_metas:
            meta.client_meta.cloud_client = client
        for meta in large_file_metas:
            meta.client_meta.cloud_client = client

        if len(small_file_metas) > 0:
            self._download_files_in_parallel(small_file_metas)
        if len(large_file_metas) > 0:
            self._download_files_in_sequential(large_file_metas)

    def _download_files_in_parallel(self, file_metas: List["SnowflakeFileMeta"]):
        """Downloads files in parallel.

        Args:
            file_metas: List of metadata for files to be downloaded.
        """
        idx = 0
        len_file_metas = len(file_metas)
        while idx < len_file_metas:
            end_of_idx = (
                idx + self._parallel
                if idx + self._parallel <= len_file_metas
                else len_file_metas
            )

            logger.debug(f"downloading files idx: {idx} to {end_of_idx}")

            target_meta = file_metas[idx:end_of_idx]
            while True:
                pool = ThreadPoolExecutor(len(target_meta))
                results: List["SnowflakeFileMeta"] = list(
                    pool.map(SnowflakeFileTransferAgent.download_one_file, target_meta)
                )
                pool.shutdown()

                # need renew AWS token?
                retry_meta: List["SnowflakeFileMeta"] = []
                for result_meta in results:
                    if result_meta.result_status in [
                        ResultStatus.RENEW_TOKEN,
                        ResultStatus.RENEW_PRESIGNED_URL,
                    ]:
                        retry_meta.append(result_meta)
                    else:
                        self._results.append(result_meta)

                if len(retry_meta) == 0:
                    # no new AWS token is required
                    break
                if any(
                    [
                        result_meta.result_status == ResultStatus.RENEW_TOKEN
                        for result_meta in results
                    ]
                ):
                    client = self.renew_expired_client()
                    for result_meta in retry_meta:
                        result_meta.client_meta.cloud_client = client
                if any(
                    [
                        result_meta.result_status == ResultStatus.RENEW_PRESIGNED_URL
                        for result_meta in results
                    ]
                ):
                    self._update_file_metas_with_presigned_url()
                if end_of_idx < len_file_metas:
                    for idx0 in range(idx + self._parallel, len_file_metas):
                        file_metas[idx0].client_meta.cloud_client = client
                target_meta = retry_meta

            if end_of_idx == len_file_metas:
                break
            idx += self._parallel

    def _download_files_in_sequential(self, file_metas: List["SnowflakeFileMeta"]):
        """Downloads files in sequential. Retry if the access token expires.

        Args:
            file_metas: List of metadata for files to be downloaded.
        """
        idx = 0
        len_file_metas = len(file_metas)
        while idx < len_file_metas:
            result = SnowflakeFileTransferAgent.download_one_file(file_metas[idx])
            if result.result_status == ResultStatus.RENEW_TOKEN:
                client = self.renew_expired_client()
                for idx0 in range(idx, len_file_metas):
                    file_metas[idx0].client_meta.cloud_client = client
                continue
            elif result.result_status == ResultStatus.RENEW_PRESIGNED_URL:
                self._update_file_metas_with_presigned_url()
                continue
            self._results.append(result)
            idx += 1
            if INJECT_WAIT_IN_PUT > 0:
                logger.debug(f"LONGEVITY TEST: waiting for {INJECT_WAIT_IN_PUT}")
                sleep(INJECT_WAIT_IN_PUT)

    @staticmethod
    def download_one_file(meta: "SnowflakeFileMeta") -> "SnowflakeFileMeta":
        """Download a one file.

        Args:
            meta: Metadata for file to be downloaded.

        Returns:
            Metadata of downloaded file.
        """
        tmp_dir = tempfile.mkdtemp()
        meta.tmp_dir = tmp_dir
        try:
            storage_client = SnowflakeFileTransferAgent.get_storage_client(
                meta.stage_location_type
            )
            storage_client.download_one_file(meta)
            logger.debug(
                f"done: status={meta.result_status}, file={meta.dst_file_name}"
            )
        except Exception as e:
            logger.exception(f"Failed to download a file: {meta.dst_file_name}")
            meta.dst_file_size = -1
            if meta.result_status is not None:
                meta.result_status = ResultStatus.ERROR
            meta.error_details = str(e)
            meta.error_details += f", file={meta.dst_file_name}"
        finally:
            logger.debug(f"cleaning up tmp dir: {tmp_dir}")
            shutil.rmtree(tmp_dir)
        return meta

    def renew_expired_client(self):
        logger.debug("renewing expired aws token")
        ret = self._cursor._execute_helper(
            self._command
        )  # rerun the command to get the credential
        stage_info = ret["data"]["stageInfo"]
        storage_client = SnowflakeFileTransferAgent.get_storage_client(
            self._stage_location_type
        )
        return storage_client.create_client(
            stage_info,
            use_accelerate_endpoint=self._use_accelerate_endpoint,
            use_s3_regional_url=self._use_s3_regional_url,
            s3_connection_pool_size=self._cursor._connection._s3_connection_pool_size,
        )

    def _update_file_metas_with_presigned_url(self):
        """Updates the file metas with presigned urls if any.

        Currently only the file metas generated for PUT/GET on a GCP account need the presigned urls.
        """
        storage_client_class = SnowflakeFileTransferAgent.get_storage_client(
            self._stage_location_type
        )

        # presigned url only applies to remote storage
        if storage_client_class is not SnowflakeRemoteStorageUtil:
            return

        storage_util_class = SnowflakeRemoteStorageUtil.get_for_storage_type(
            self._stage_location_type
        )

        # presigned url only applies to GCS
        if storage_util_class in [SnowflakeGCSUtil]:
            if self._command_type == CMD_TYPE_UPLOAD:
                logger.debug("getting presigned urls for upload")

                # Rewrite the command such that a new PUT call is made for each file
                # represented by the regex (if present) separately. This is the only
                # way to get the presigned url for that file.
                file_path_to_be_replaced = self.get_local_file_path_from_put_command(
                    self._command
                )

                for m in self._file_metadata:
                    # At this point the connector has already figured out and
                    # validated that the local file exists and has also decided
                    # upon the destination file name and the compression type.
                    # The only thing that's left to do is to get the presigned
                    # url for the destination file. If the command originally
                    # referred to a single file, then the presigned url got in
                    # that case is simply ignore, since the file name is not what
                    # we want.

                    # GS only looks at the file name at the end of local file
                    # path to figure out the remote object name. Hence the prefix
                    # for local path is not necessary in the reconstructed command.
                    file_path_to_replace_with = m.dst_file_name
                    command_with_single_file = self._command
                    command_with_single_file = command_with_single_file.replace(
                        file_path_to_be_replaced, file_path_to_replace_with
                    )

                    logger.debug(
                        "getting presigned url for %s", file_path_to_replace_with
                    )

                    ret = self._cursor._execute_helper(command_with_single_file)

                    if ret.get("data", dict()).get("stageInfo"):
                        m.stage_info = ret["data"]["stageInfo"]
                        m.presigned_url = m.stage_info.get("presignedUrl")
            elif self._command_type == CMD_TYPE_DOWNLOAD:
                logger.debug("updating download file metas with presigned urls")

                for idx, m in enumerate(self._file_metadata):
                    m.presigned_url = (
                        self._presigned_urls[idx]
                        if len(self._presigned_urls) > idx
                        else None
                    )

    def result(self):
        converter_class = self._cursor._connection.converter_class
        rowset = []
        if self._command_type == CMD_TYPE_UPLOAD:
            if hasattr(self, "_results"):
                for meta in self._results:
                    if meta.src_compression_type is not None:
                        src_compression_type = meta.src_compression_type.name
                    else:
                        src_compression_type = "NONE"

                    if meta.dst_compression_type is not None:
                        dst_compression_type = meta.dst_compression_type.name
                    else:
                        dst_compression_type = "NONE"

                    error_details = meta.error_details or ""

                    src_file_size = (
                        meta.src_file_size
                        if converter_class != SnowflakeConverterSnowSQL
                        else str(meta.src_file_size)
                    )

                    dst_file_size = (
                        meta.dst_file_size
                        if converter_class != SnowflakeConverterSnowSQL
                        else str(meta.dst_file_size)
                    )

                    logger.debug(
                        "raise_put_get_error: %s, %s, %s, %s, %s",
                        self._raise_put_get_error,
                        meta.result_status,
                        type(meta.result_status),
                        ResultStatus.ERROR,
                        type(ResultStatus.ERROR),
                    )
                    if self._raise_put_get_error and error_details:
                        Error.errorhandler_wrapper(
                            self._cursor.connection,
                            self._cursor,
                            OperationalError,
                            {
                                "msg": error_details,
                                "errno": ER_FAILED_TO_UPLOAD_TO_STAGE,
                            },
                        )
                    rowset.append(
                        [
                            meta.name,
                            meta.dst_file_name,
                            src_file_size,
                            dst_file_size,
                            src_compression_type,
                            dst_compression_type,
                            meta.result_status,
                            error_details,
                        ]
                    )
            return {
                "rowtype": [
                    result_text_column_desc("source"),
                    result_text_column_desc("target"),
                    result_fixed_column_desc("source_size"),
                    result_fixed_column_desc("target_size"),
                    result_text_column_desc("source_compression"),
                    result_text_column_desc("target_compression"),
                    result_text_column_desc("status"),
                    result_text_column_desc("message"),
                ],
                "rowset": sorted(rowset),
            }
        else:  # DOWNLOAD
            if hasattr(self, "_results"):
                for meta in self._results:
                    dst_file_size = (
                        meta.dst_file_size
                        if converter_class != SnowflakeConverterSnowSQL
                        else str(meta.dst_file_size)
                    )

                    error_details = meta.error_details or ""

                    if self._raise_put_get_error and error_details:
                        Error.errorhandler_wrapper(
                            self._cursor.connection,
                            self._cursor,
                            OperationalError,
                            {
                                "msg": error_details,
                                "errno": ER_FAILED_TO_DOWNLOAD_FROM_STAGE,
                            },
                        )

                    rowset.append(
                        [
                            meta.dst_file_name,
                            dst_file_size,
                            meta.result_status,
                            error_details,
                        ]
                    )
            return {
                "rowtype": [
                    result_text_column_desc("file"),
                    result_fixed_column_desc("size"),
                    result_text_column_desc("status"),
                    result_text_column_desc("message"),
                ],
                "rowset": sorted(rowset),
            }

    def _expand_filenames(self, locations):
        canonical_locations = []
        for file_name in locations:
            if self._command_type == CMD_TYPE_UPLOAD:
                file_name = os.path.expanduser(file_name)
                if not os.path.isabs(file_name):
                    file_name = os.path.join(GET_CWD(), file_name)
                if (
                    IS_WINDOWS
                    and len(file_name) > 2
                    and file_name[0] == "/"
                    and file_name[2] == ":"
                ):
                    # Windows path: /C:/data/file1.txt where it starts with slash
                    # followed by a drive letter and colon.
                    file_name = file_name[1:]
                files = glob.glob(file_name)
                canonical_locations += files
            else:
                canonical_locations.append(file_name)

        return canonical_locations

    def _init_encryption_material(self):
        self._encryption_material = []

        if (
            "data" in self._ret
            and "encryptionMaterial" in self._ret["data"]
            and self._ret["data"]["encryptionMaterial"] is not None
        ):
            root_node = self._ret["data"]["encryptionMaterial"]
            logger.debug(self._command_type)

            if self._command_type == CMD_TYPE_UPLOAD:
                self._encryption_material.append(
                    SnowflakeFileEncryptionMaterial(
                        query_stage_master_key=root_node["queryStageMasterKey"],
                        query_id=root_node["queryId"],
                        smk_id=root_node["smkId"],
                    )
                )
            else:
                for elem in root_node:
                    if elem is not None:
                        self._encryption_material.append(
                            SnowflakeFileEncryptionMaterial(
                                query_stage_master_key=elem["queryStageMasterKey"],
                                query_id=elem["queryId"],
                                smk_id=elem["smkId"],
                            )
                        )

    def _parse_command(self):
        if "data" in self._ret:
            self._command_type = self._ret["data"]["command"]
        else:
            self._command_type = "Unknown"

        self._init_encryption_material()
        if (
            "data" in self._ret
            and "src_locations" in self._ret["data"]
            and isinstance(self._ret["data"]["src_locations"], list)
        ):
            self._src_locations = self._ret["data"]["src_locations"]
        else:
            Error.errorhandler_wrapper(
                self._cursor.connection,
                self._cursor,
                DatabaseError,
                {
                    "msg": "Failed to parse the location",
                    "errno": ER_INVALID_STAGE_LOCATION,
                },
            )

        if self._command_type == CMD_TYPE_UPLOAD:
            if self._source_from_stream:
                self._src_files = self._src_locations
            else:
                self._src_files = list(self._expand_filenames(self._src_locations))
            self._auto_compress = (
                "autoCompress" not in self._ret["data"]
                or self._ret["data"]["autoCompress"]
            )
            self._source_compression = (
                self._ret["data"]["sourceCompression"].lower()
                if "sourceCompression" in self._ret["data"]
                else ""
            )
        else:
            self._src_files = list(self._src_locations)
            self._src_file_to_encryption_material = {}
            if len(self._ret["data"]["src_locations"]) == len(
                self._encryption_material
            ):
                for idx, src_file in enumerate(self._src_files):
                    logger.debug(src_file)
                    self._src_file_to_encryption_material[
                        src_file
                    ] = self._encryption_material[idx]
            elif len(self._encryption_material) != 0:
                # some encryption material exists. Zero means no encryption
                Error.errorhandler_wrapper(
                    self._cursor.connection,
                    self._cursor,
                    InternalError,
                    {
                        "msg": (
                            "The number of downloading files doesn't match "
                            f"the encryption materials: files={len(self._ret['data']['src_locations'])}, "
                            f"encmat={len(self._encryption_material)}"
                        ),
                        "errno": ER_INTERNAL_NOT_MATCH_ENCRYPT_MATERIAL,
                    },
                )

            self._local_location = os.path.expanduser(
                self._ret["data"]["localLocation"]
            )
            if not os.path.isdir(self._local_location):
                # NOTE: isdir follows the symlink
                Error.errorhandler_wrapper(
                    self._cursor.connection,
                    self._cursor,
                    ProgrammingError,
                    {
                        "msg": "The local path is not a directory: "
                        f"{self._local_location}",
                        "errno": ER_LOCAL_PATH_NOT_DIRECTORY,
                    },
                )

        self._parallel = self._ret["data"].get("parallel", 1)
        self._overwrite = self._force_put_overwrite or self._ret["data"].get(
            "overwrite", False
        )
        self._stage_location_type = self._ret["data"]["stageInfo"][
            "locationType"
        ].upper()
        self._stage_location = self._ret["data"]["stageInfo"]["location"]
        self._stage_info = self._ret["data"]["stageInfo"]
        self._presigned_urls = self._ret["data"].get("presignedUrls")

        if self.get_storage_client(self._stage_location_type) is None:
            Error.errorhandler_wrapper(
                self._cursor.connection,
                self._cursor,
                OperationalError,
                {
                    "msg": f"Destination location type is not valid: {self._stage_location_type}",
                    "errno": ER_INVALID_STAGE_FS,
                },
            )

    def _init_file_metadata(self):
        logger.debug(f"command type: {self._command_type}")

        if self._command_type == CMD_TYPE_UPLOAD:
            if len(self._src_files) == 0:
                file_name = (
                    self._ret["data"]["src_locations"]
                    if "data" in self._ret and "src_locations" in self._ret["data"]
                    else "None"
                )
                Error.errorhandler_wrapper(
                    self._cursor.connection,
                    self._cursor,
                    ProgrammingError,
                    {
                        "msg": f"File doesn't exist: {file_name}",
                        "errno": ER_FILE_NOT_EXISTS,
                    },
                )
            if not self._source_from_stream:
                for file_name in self._src_files:
                    if not os.path.exists(file_name):
                        Error.errorhandler_wrapper(
                            self._cursor.connection,
                            self._cursor,
                            ProgrammingError,
                            {
                                "msg": f"File doesn't exist: {file_name}",
                                "errno": ER_FILE_NOT_EXISTS,
                            },
                        )
                    elif os.path.isdir(file_name):
                        Error.errorhandler_wrapper(
                            self._cursor.connection,
                            self._cursor,
                            ProgrammingError,
                            {
                                "msg": f"Not a file but a directory: {file_name}",
                                "errno": ER_FILE_NOT_EXISTS,
                            },
                        )
                    statinfo = os.stat(file_name)
                    self._file_metadata.append(
                        SnowflakeFileMeta(
                            name=os.path.basename(file_name),
                            src_file_name=file_name,
                            src_file_size=statinfo.st_size,
                            stage_location_type=self._stage_location_type,
                            client_meta=SFResourceMeta(
                                stage_info=self._stage_info,
                                use_s3_regional_url=self._use_s3_regional_url,
                                use_accelerate_endpoint=self._use_accelerate_endpoint,
                            ),
                            encryption_material=self._encryption_material[0]
                            if len(self._encryption_material) > 0
                            else None,
                        )
                    )
            else:
                file_name = self._src_files[0]
                self._file_metadata.append(
                    SnowflakeFileMeta(
                        name=os.path.basename(file_name),
                        src_file_name=file_name,
                        src_stream=self._source_from_stream,
                        src_file_size=self._source_from_stream.seek(0, os.SEEK_END),
                        stage_location_type=self._stage_location_type,
                        client_meta=SFResourceMeta(
                            stage_info=self._stage_info,
                            use_s3_regional_url=self._use_s3_regional_url,
                            use_accelerate_endpoint=self._use_accelerate_endpoint,
                        ),
                        encryption_material=self._encryption_material[0]
                        if len(self._encryption_material) > 0
                        else None,
                    )
                )
                self._source_from_stream.seek(0)
        elif self._command_type == CMD_TYPE_DOWNLOAD:
            for file_name in self._src_files:
                if len(file_name) > 0:
                    logger.debug(file_name)
                    first_path_sep = file_name.find("/")
                    dst_file_name = (
                        file_name[first_path_sep + 1 :]
                        if first_path_sep >= 0
                        else file_name
                    )
                    self._file_metadata.append(
                        SnowflakeFileMeta(
                            name=os.path.basename(file_name),
                            src_file_name=file_name,
                            dst_file_name=dst_file_name,
                            stage_location_type=self._stage_location_type,
                            client_meta=SFResourceMeta(
                                stage_info=self._stage_info,
                                use_s3_regional_url=self._use_s3_regional_url,
                                use_accelerate_endpoint=self._use_accelerate_endpoint,
                            ),
                            local_location=self._local_location,
                            encryption_material=self._src_file_to_encryption_material[
                                file_name
                            ]
                            if file_name in self._src_file_to_encryption_material
                            else None,
                        )
                    )

    def _process_file_compression_type(self):
        user_specified_source_compression = None
        if self._source_compression == "auto_detect":
            auto_detect = True
        elif self._source_compression == "none":
            auto_detect = False
        else:
            user_specified_source_compression: Optional[
                "CompressionType"
            ] = lookup_by_mime_sub_type(self._source_compression)
            if (
                user_specified_source_compression is None
                or not user_specified_source_compression.is_supported
            ):
                Error.errorhandler_wrapper(
                    self._cursor.connection,
                    self._cursor,
                    ProgrammingError,
                    {
                        "msg": f"Feature is not supported: {user_specified_source_compression}",
                        "errno": ER_COMPRESSION_NOT_SUPPORTED,
                    },
                )

            auto_detect = False

        for m in self._file_metadata:
            file_name = m.src_file_name

            current_file_compression_type: Optional["CompressionType"] = None
            if auto_detect:
                mimetypes.init()
                _, encoding = mimetypes.guess_type(file_name)

                if encoding is None:
                    test = None
                    if not self._source_from_stream:
                        with open(file_name, "rb") as f:
                            test = f.read(4)
                    else:
                        test = self._source_from_stream.read(4)
                        self._source_from_stream.seek(0)
                    if file_name.endswith(".br"):
                        encoding = "br"
                    elif test and test[:3] == b"ORC":
                        encoding = "orc"
                    elif test and test == b"PAR1":
                        encoding = "parquet"
                    elif test and (int(binascii.hexlify(test), 16) == 0x28B52FFD):
                        encoding = "zstd"

                if encoding is not None:
                    logger.debug(
                        "detected the encoding %s: file=%s", encoding, file_name
                    )
                    current_file_compression_type = lookup_by_mime_sub_type(encoding)
                else:
                    logger.debug("no file encoding was detected: file=%s", file_name)

                if (
                    current_file_compression_type is not None
                    and not current_file_compression_type.is_supported
                ):
                    Error.errorhandler_wrapper(
                        self._cursor.connection,
                        self._cursor,
                        ProgrammingError,
                        {
                            "msg": f"Feature is not supported: {current_file_compression_type}",
                            "errno": ER_COMPRESSION_NOT_SUPPORTED,
                        },
                    )
            else:
                current_file_compression_type = user_specified_source_compression

            if current_file_compression_type is not None:
                m.src_compression_type = current_file_compression_type
                if current_file_compression_type.is_supported:
                    m.dst_compression_type = current_file_compression_type
                    m.require_compress = False
                    m.dst_file_name = m.name
                else:
                    Error.errorhandler_wrapper(
                        self._cursor.connection,
                        self._cursor,
                        ProgrammingError,
                        {
                            "msg": f"Feature is not supported: {current_file_compression_type}",
                            "errno": ER_COMPRESSION_NOT_SUPPORTED,
                        },
                    )
            else:
                # src is not compressed but the destination want to be
                # compressed unless the users disable it
                m.require_compress = self._auto_compress
                m.src_compression_type = None
                if self._auto_compress:
                    m.dst_file_name = m.name + CompressionTypes["GZIP"].file_extension
                    m.dst_compression_type = CompressionTypes["GZIP"]
                else:
                    m.dst_file_name = m.name
                    m.dst_compression_type = None

    def get_local_file_path_from_put_command(self, command):
        """Get the local file path from PUT command (Logic adopted from JDBC, written by Polita).

        Args:
            command: Command to be parsed and get the local file path out of.

        Returns:
            The local file path.
        """
        if command is None or len(command) == 0 or FILE_PROTOCOL not in command:
            return None

        if not self._cursor.PUT_SQL_RE.match(command):
            return None

        file_path_begin_index = command.find(FILE_PROTOCOL)
        is_file_path_quoted = command[file_path_begin_index - 1] == "'"
        file_path_begin_index += len(FILE_PROTOCOL)

        file_path = ""

        if is_file_path_quoted:
            file_path_end_index = command.find("'", file_path_begin_index)

            if file_path_end_index > file_path_begin_index:
                file_path = command[file_path_begin_index:file_path_end_index]
        else:
            index_list = []
            for delimiter in [" ", "\n", ";"]:
                index = command.find(delimiter, file_path_begin_index)
                if index != -1:
                    index_list += [index]

            file_path_end_index = min(index_list) if index_list else -1

            if file_path_end_index > file_path_begin_index:
                file_path = command[file_path_begin_index:file_path_end_index]
            elif file_path_end_index == -1:
                file_path = command[file_path_begin_index:]

        return file_path



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2021 Snowflake Computing Inc. All right reserved.
#

from __future__ import division

import os
import shutil
import time
from io import BytesIO
from logging import getLogger
from typing import TYPE_CHECKING, NamedTuple

from boto3 import Session

from .azure_util_sdk import SnowflakeAzureUtil
from .constants import ResultStatus
from .encryption_util import SnowflakeEncryptionUtil
from .gcs_util_sdk import SnowflakeGCSUtil
from .s3_util_sdk import SnowflakeS3Util

if TYPE_CHECKING:  # pragma: no cover
    from .file_transfer_agent_sdk import SnowflakeFileMeta

DEFAULT_CONCURRENCY = 1
DEFAULT_MAX_RETRY = 5

logger = getLogger(__name__)


class SnowflakeFileEncryptionMaterial(NamedTuple):
    query_stage_master_key: str  # query stage master key
    query_id: str  # query id
    smk_id: int  # SMK id


class NeedRenewTokenError(Exception):
    pass


class SnowflakeRemoteStorageUtil(object):
    @staticmethod
    def get_for_storage_type(_type):
        if _type == "S3":
            return SnowflakeS3Util
        elif _type == "AZURE":
            return SnowflakeAzureUtil
        elif _type == "GCS":
            return SnowflakeGCSUtil
        else:
            return None

    @staticmethod
    def create_client(
        stage_info,
        use_accelerate_endpoint: bool = False,
        use_s3_regional_url: bool = False,
        s3_connection_pool_size: int = 1,
    ) -> Session.resource:
        util_class = SnowflakeRemoteStorageUtil.get_for_storage_type(
            stage_info["locationType"]
        )
        return util_class.create_client(
            stage_info,
            use_accelerate_endpoint=use_accelerate_endpoint,
            use_s3_regional_url=use_s3_regional_url,
            s3_connection_pool_size=s3_connection_pool_size,
        )

    @staticmethod
    def upload_one_file(meta: "SnowflakeFileMeta") -> None:
        """Uploads a file to S3."""
        encryption_metadata = None

        if meta.encryption_material is not None:
            if meta.src_stream is None:
                (encryption_metadata, data_file) = SnowflakeEncryptionUtil.encrypt_file(
                    meta.encryption_material,
                    meta.real_src_file_name,
                    tmp_dir=meta.tmp_dir,
                )
                logger.debug(
                    f"encrypted data file={data_file}, size={os.path.getsize(data_file)}"
                )
            else:
                encrypted_stream = BytesIO()
                src_stream = meta.real_src_stream or meta.src_stream
                src_stream.seek(0)
                encryption_metadata = SnowflakeEncryptionUtil.encrypt_stream(
                    meta.encryption_material, src_stream, encrypted_stream
                )
                src_stream.seek(0)
                logger.debug(
                    f"encrypted data stream size={encrypted_stream.seek(0, os.SEEK_END)}"
                )
                encrypted_stream.seek(0)
                if meta.real_src_stream is not None:
                    meta.real_src_stream.close()
                meta.real_src_stream = encrypted_stream
                data_file = meta.real_src_file_name
        else:
            logger.debug("not encrypted data file")
            data_file = meta.real_src_file_name

        util_class = SnowflakeRemoteStorageUtil.get_for_storage_type(
            meta.client_meta.stage_info["locationType"]
        )

        logger.debug(
            f"putting a file: {meta.client_meta.stage_info['location']}, {meta.dst_file_name}"
        )

        max_concurrency = meta.parallel
        last_err = None
        max_retry = DEFAULT_MAX_RETRY
        for retry in range(max_retry):
            if not meta.overwrite:
                file_header = util_class.get_file_header(meta, meta.dst_file_name)

                if file_header and meta.result_status == ResultStatus.UPLOADED:
                    logger.debug(
                        f'file already exists location="{meta.client_meta.stage_info["location"]}", '
                        f'file_name="{meta.dst_file_name}"'
                    )
                    meta.dst_file_size = 0
                    meta.result_status = ResultStatus.SKIPPED
                    return

            if meta.overwrite or meta.result_status == ResultStatus.NOT_FOUND_FILE:
                util_class.upload_file(
                    data_file,
                    meta,
                    encryption_metadata,
                    max_concurrency,
                    multipart_threshold=meta.multipart_threshold,
                )

            if meta.result_status == ResultStatus.UPLOADED:
                return
            elif meta.result_status == ResultStatus.RENEW_TOKEN:
                return
            elif meta.result_status == ResultStatus.RENEW_PRESIGNED_URL:
                return
            elif meta.result_status == ResultStatus.NEED_RETRY:
                last_err = meta.last_error
                logger.debug(
                    f"Failed to upload a file: {data_file}, err: {last_err}. Retrying with "
                    f"max concurrency: {max_concurrency}"
                )
                if not meta.no_sleeping_time:
                    sleeping_time = min(2 ** retry, 16)
                    logger.debug(f"sleeping {sleeping_time}")
                    time.sleep(sleeping_time)
            elif meta.result_status == ResultStatus.NEED_RETRY_WITH_LOWER_CONCURRENCY:
                last_err = meta.last_error
                max_concurrency = meta.parallel - int(retry * meta.parallel / max_retry)
                max_concurrency = max(DEFAULT_CONCURRENCY, max_concurrency)
                meta.last_max_concurrency = max_concurrency

                logger.debug(
                    f"Failed to upload a file: {data_file}, err: {last_err}. Retrying with "
                    f"max concurrency: {max_concurrency}"
                )
                if meta.no_sleeping_time is None:
                    sleeping_time = min(2 ** retry, 16)
                    logger.debug(f"sleeping: {sleeping_time}")
                    time.sleep(sleeping_time)
        else:
            if last_err:
                raise last_err
            else:
                msg = f"Unknown Error in uploading a file: {data_file}"
                raise Exception(msg)

    @staticmethod
    def download_one_file(meta: "SnowflakeFileMeta") -> None:
        """Downloads a file from S3."""
        full_dst_file_name = os.path.join(
            meta.local_location, os.path.basename(meta.dst_file_name)
        )
        full_dst_file_name = os.path.realpath(full_dst_file_name)
        # TODO: validate full_dst_file_name is under the writable directory
        base_dir = os.path.dirname(full_dst_file_name)
        if not os.path.exists(base_dir):
            os.makedirs(base_dir)

        util_class = SnowflakeRemoteStorageUtil.get_for_storage_type(
            meta.client_meta.stage_info["locationType"]
        )
        file_header = util_class.get_file_header(meta, meta.src_file_name)

        if file_header:
            meta.src_file_size = file_header.content_length

        full_dst_file_name = os.path.join(
            meta.local_location, os.path.basename(meta.dst_file_name)
        )
        full_dst_file_name = os.path.realpath(full_dst_file_name)

        max_concurrency = meta.parallel
        last_err = None
        max_retry = DEFAULT_MAX_RETRY
        for retry in range(max_retry):
            util_class._native_download_file(meta, full_dst_file_name, max_concurrency)
            if meta.result_status == ResultStatus.DOWNLOADED:
                if meta.encryption_material is not None:
                    logger.debug(f"encrypted data file={full_dst_file_name}")

                    # For storage utils that do not have the privilege of
                    # getting the metadata early, both object and metadata
                    # are downloaded at once. In which case, the file meta will
                    # be updated with all the metadata that we need and
                    # then we can call get_file_header to get just that and also
                    # preserve the idea of getting metadata in the first place.
                    # One example of this is the utils that use presigned url
                    # for upload/download and not the storage client library.
                    if meta.presigned_url is not None:
                        file_header = util_class.get_file_header(
                            meta, meta.src_file_name
                        )

                    tmp_dst_file_name = SnowflakeEncryptionUtil.decrypt_file(
                        file_header.encryption_metadata,
                        meta.encryption_material,
                        full_dst_file_name,
                        tmp_dir=meta.tmp_dir,
                    )
                    shutil.copyfile(tmp_dst_file_name, full_dst_file_name)
                    os.unlink(tmp_dst_file_name)
                else:
                    logger.debug(f"not encrypted data file={full_dst_file_name}")

                stat_info = os.stat(full_dst_file_name)
                meta.dst_file_size = stat_info.st_size
                return
            elif meta.result_status == ResultStatus.RENEW_PRESIGNED_URL:
                return
            elif meta.result_status == ResultStatus.RENEW_TOKEN:
                return
            elif meta.result_status == ResultStatus.NEED_RETRY_WITH_LOWER_CONCURRENCY:
                max_concurrency = meta.parallel - int(retry * meta.parallel / max_retry)
                max_concurrency = max(DEFAULT_CONCURRENCY, max_concurrency)
                meta.last_max_concurrency = max_concurrency
                last_err = meta.last_error
                logger.debug(
                    f"Failed to download a file: {full_dst_file_name}, err: {last_err}. Retrying with "
                    f"max concurrency: {max_concurrency}"
                )
                if not meta.no_sleeping_time:
                    sleeping_time = min(2 ** retry, 16)
                    logger.debug("sleeping: %s", sleeping_time)
                    time.sleep(sleeping_time)
            elif meta.result_status == ResultStatus.NEED_RETRY:
                last_err = meta.last_error
                logger.debug(
                    f"Failed to download a file: {full_dst_file_name}, err: {last_err}. Retrying with "
                    f"max concurrency: {max_concurrency}"
                )
                if not meta.no_sleeping_time:
                    sleeping_time = min(2 ** retry, 16)
                    logger.debug(f"sleeping: {sleeping_time}")
                    time.sleep(sleeping_time)
        else:
            if last_err:
                raise last_err
            else:
                msg = f"Unknown Error in downloading a file: {full_dst_file_name}"
                raise Exception(msg)

    @staticmethod
    def upload_one_file_with_retry(meta: "SnowflakeFileMeta") -> None:
        """Uploads one file with retry."""
        util_class = SnowflakeRemoteStorageUtil.get_for_storage_type(
            meta.client_meta.stage_info["locationType"]
        )
        for _ in range(10):
            # retry
            SnowflakeRemoteStorageUtil.upload_one_file(meta)
            if meta.result_status == ResultStatus.UPLOADED:
                for _ in range(10):
                    util_class.get_file_header(meta, meta.dst_file_name)
                    if meta.result_status == ResultStatus.NOT_FOUND_FILE:
                        time.sleep(1)  # wait 1 second
                        logger.debug("not found. double checking...")
                        continue
                    break
                else:
                    # not found. retry with the outer loop
                    logger.debug("not found. gave up. re-uploading...")
                    continue
            break
        else:
            # could not upload a file even after retry
            meta.result_status = ResultStatus.ERROR



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2021 Snowflake Computing Inc. All right reserved.
#

from .converter import SnowflakeConverter


class SnowflakeNoConverterToPython(SnowflakeConverter):
    def __init__(self, **kwargs):
        super(SnowflakeNoConverterToPython, self).__init__(**kwargs)

    def to_python_method(self, type_name, column):
        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2021 Snowflake Computing Inc. All right reserved.
#

import logging
import re
import signal
import sys
import time
import uuid
from enum import Enum
from logging import getLogger
from threading import Lock, Timer
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generator,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from snowflake.connector.result_batch import create_batches_from_response
from snowflake.connector.result_set import ResultSet

from .bind_upload_agent import BindUploadAgent, BindUploadError
from .compat import BASE_EXCEPTION_CLASS
from .constants import (
    FIELD_NAME_TO_ID,
    PARAMETER_PYTHON_CONNECTOR_QUERY_RESULT_FORMAT,
    FileTransferType,
    QueryStatus,
)
from .errorcode import (
    ER_CURSOR_IS_CLOSED,
    ER_FAILED_TO_REWRITE_MULTI_ROW_INSERT,
    ER_INVALID_VALUE,
    ER_NO_ARROW_RESULT,
    ER_NO_PYARROW,
    ER_NO_PYARROW_SNOWSQL,
    ER_NOT_POSITIVE_SIZE,
    ER_UNSUPPORTED_METHOD,
)
from .errors import (
    DatabaseError,
    Error,
    InterfaceError,
    NotSupportedError,
    ProgrammingError,
)
from .file_transfer_agent import SnowflakeFileTransferAgent
from .file_transfer_agent_sdk import (
    SnowflakeFileTransferAgent as SnowflakeFileTransferAgentSdk,
)
from .options import installed_pandas, pandas
from .sqlstate import SQLSTATE_FEATURE_NOT_SUPPORTED
from .telemetry import TelemetryData, TelemetryField
from .time_util import get_time_millis

if TYPE_CHECKING:  # pragma: no cover
    from .connection import SnowflakeConnection
    from .file_transfer_agent import SnowflakeProgressPercentage
    from .result_batch import ResultBatch


logger = getLogger(__name__)

if installed_pandas:
    from pyarrow import Table
else:
    logger.debug("Failed to import pyarrow. Cannot use pandas fetch API")
    Table = None

try:
    from .arrow_iterator import PyArrowIterator  # NOQA

    CAN_USE_ARROW_RESULT_FORMAT = True
except ImportError as e:  # pragma: no cover
    logger.debug(
        f"Failed to import ArrowResult. No Apache Arrow result set format can be used. ImportError: {e}",
    )
    CAN_USE_ARROW_RESULT_FORMAT = False

STATEMENT_TYPE_ID_DML = 0x3000
STATEMENT_TYPE_ID_INSERT = STATEMENT_TYPE_ID_DML + 0x100
STATEMENT_TYPE_ID_UPDATE = STATEMENT_TYPE_ID_DML + 0x200
STATEMENT_TYPE_ID_DELETE = STATEMENT_TYPE_ID_DML + 0x300
STATEMENT_TYPE_ID_MERGE = STATEMENT_TYPE_ID_DML + 0x400
STATEMENT_TYPE_ID_MULTI_TABLE_INSERT = STATEMENT_TYPE_ID_DML + 0x500

STATEMENT_TYPE_ID_DML_SET = frozenset(
    [
        STATEMENT_TYPE_ID_DML,
        STATEMENT_TYPE_ID_INSERT,
        STATEMENT_TYPE_ID_UPDATE,
        STATEMENT_TYPE_ID_DELETE,
        STATEMENT_TYPE_ID_MERGE,
        STATEMENT_TYPE_ID_MULTI_TABLE_INSERT,
    ]
)

DESC_TABLE_RE = re.compile(r"desc(?:ribe)?\s+([\w_]+)\s*;?\s*$", flags=re.IGNORECASE)

LOG_MAX_QUERY_LENGTH = 80

ASYNC_NO_DATA_MAX_RETRY = 24
ASYNC_RETRY_PATTERN = [1, 1, 2, 3, 4, 8, 10]
INCIDENT_BLACKLIST = (KeyError, ValueError, TypeError)


class ResultMetadata(NamedTuple):
    name: str
    type_code: int
    display_size: int
    internal_size: int
    precision: int
    scale: int
    is_nullable: bool

    @classmethod
    def from_column(cls, col: Dict[str, Any]):
        """Initializes a ResultMetadata object from the column description in the query response."""
        return cls(
            col["name"],
            FIELD_NAME_TO_ID[col["type"].upper()],
            None,
            col["length"],
            col["precision"],
            col["scale"],
            col["nullable"],
        )


# TODO: once we drop 3.6 support the return type becomes NoReturn
def exit_handler(*_) -> None:  # pragma: no cover
    """Handler for signal. When called, it will raise SystemExit with exit code FORCE_EXIT."""
    print("\nForce exit")
    logger.info("Force exit")
    sys.exit(1)


class ResultState(Enum):
    DEFAULT = 1
    VALID = 2
    RESET = 3


class SnowflakeCursor:
    """Implementation of Cursor object that is returned from Connection.cursor() method.

    Attributes:
        description: A list of namedtuples about metadata for all columns.
        rowcount: The number of records updated or selected. If not clear, -1 is returned.
        rownumber: The current 0-based index of the cursor in the result set or None if the index cannot be
            determined.
        sfqid: Snowflake query id in UUID form. Include this in the problem report to the customer support.
        sqlstate: Snowflake SQL State code.
        timestamp_output_format: Snowflake timestamp_output_format for timestamps.
        timestamp_ltz_output_format: Snowflake output format for LTZ timestamps.
        timestamp_tz_output_format: Snowflake output format for TZ timestamps.
        timestamp_ntz_output_format: Snowflake output format for NTZ timestamps.
        date_output_format: Snowflake output format for dates.
        time_output_format: Snowflake output format for times.
        timezone: Snowflake timezone.
        binary_output_format: Snowflake output format for binary fields.
        arraysize: The default number of rows fetched by fetchmany.
        connection: The connection object by which the cursor was created.
        errorhandle: The class that handles error handling.
        is_file_transfer: Whether, or not the current command is a put, or get.

    TODO:
        Most of these attributes have no reason to be properties, we could just store them in public variables.
        Calling a function is expensive in Python and most of these getters are unnecessary.
    """

    PUT_SQL_RE = re.compile(r"^(?:/\*.*\*/\s*)*put\s+", flags=re.IGNORECASE)
    GET_SQL_RE = re.compile(r"^(?:/\*.*\*/\s*)*get\s+", flags=re.IGNORECASE)
    INSERT_SQL_RE = re.compile(r"^insert\s+into", flags=re.IGNORECASE)
    COMMENT_SQL_RE = re.compile(r"/\*.*\*/")
    INSERT_SQL_VALUES_RE = re.compile(
        r".*VALUES\s*(\(.*\)).*", re.IGNORECASE | re.MULTILINE | re.DOTALL
    )
    ALTER_SESSION_RE = re.compile(
        r"alter\s+session\s+set\s+(.*)=\'?([^\']+)\'?\s*;",
        flags=re.IGNORECASE | re.MULTILINE | re.DOTALL,
    )

    @staticmethod
    def get_file_transfer_type(sql: str) -> Optional[FileTransferType]:
        """Decide whether a SQL is a file transfer and return its type.

        None is returned if the SQL isn't a file transfer so that this function can be
        used in an if-statement.
        """
        if SnowflakeCursor.PUT_SQL_RE.match(sql):
            return FileTransferType.PUT
        elif SnowflakeCursor.GET_SQL_RE.match(sql):
            return FileTransferType.GET
        return None

    def __init__(
        self,
        connection: "SnowflakeConnection",
        use_dict_result: bool = False,
    ) -> None:
        """Inits a SnowflakeCursor with a connection.

        Args:
            connection: The connection that created this cursor.
            use_dict_result: Decides whether to use dict result or not.
        """
        self._connection: "SnowflakeConnection" = connection

        self._errorhandler: Callable[
            ["SnowflakeConnection", "SnowflakeCursor", Type["Error"], Dict[str, str]],
            None,
        ] = Error.default_errorhandler
        self.messages: List[
            Tuple[Union[Type["Error"], Type[Exception]], Dict[str, Union[str, bool]]]
        ] = []
        self._timebomb: Optional[Timer] = None  # must be here for abort_exit method
        self._description: Optional[List[ResultMetadata]] = None
        self._column_idx_to_name = None
        self._sfqid = None
        self._sqlstate = None
        self._total_rowcount = -1
        self._sequence_counter = -1
        self._request_id = None
        self._is_file_transfer = False

        self._timestamp_output_format = None
        self._timestamp_ltz_output_format = None
        self._timestamp_ntz_output_format = None
        self._timestamp_tz_output_format = None
        self._date_output_format = None
        self._time_output_format = None
        self._timezone = None
        self._binary_output_format = None
        self._result: Optional[Union[Iterator[Tuple], Iterator[Dict]]] = None
        self._result_set: Optional["ResultSet"] = None
        self._result_state: ResultState = ResultState.DEFAULT
        self._use_dict_result = use_dict_result
        # TODO: self._query_result_format could be defined as an enum
        self._query_result_format: Optional[str] = None

        self._arraysize = 1  # PEP-0249: defaults to 1

        self._lock_canceling = Lock()

        self._first_chunk_time = None

        self._log_max_query_length = connection.log_max_query_length
        self._inner_cursor: Optional["SnowflakeCursor"] = None
        self._prefetch_hook = None
        self._rownumber: Optional[int] = None

        self.reset()

    def __del__(self) -> None:  # pragma: no cover
        try:
            self.close()
        except BASE_EXCEPTION_CLASS as e:
            if logger.getEffectiveLevel() <= logging.INFO:
                logger.info(e)

    @property
    def description(self) -> List[ResultMetadata]:
        return self._description

    @property
    def rowcount(self):
        return self._total_rowcount if self._total_rowcount >= 0 else None

    @property
    def rownumber(self):
        return self._rownumber if self._rownumber >= 0 else None

    @property
    def sfqid(self):
        return self._sfqid

    @property
    def sqlstate(self):
        return self._sqlstate

    @property
    def timestamp_output_format(self):
        return self._timestamp_output_format

    @property
    def timestamp_ltz_output_format(self):
        return (
            self._timestamp_ltz_output_format
            if self._timestamp_ltz_output_format
            else self._timestamp_output_format
        )

    @property
    def timestamp_tz_output_format(self):
        return (
            self._timestamp_tz_output_format
            if self._timestamp_tz_output_format
            else self._timestamp_output_format
        )

    @property
    def timestamp_ntz_output_format(self):
        return (
            self._timestamp_ntz_output_format
            if self._timestamp_ntz_output_format
            else self._timestamp_output_format
        )

    @property
    def date_output_format(self):
        return self._date_output_format

    @property
    def time_output_format(self):
        return self._time_output_format

    @property
    def timezone(self):
        return self._timezone

    @property
    def binary_output_format(self):
        return self._binary_output_format

    @property
    def arraysize(self):
        return self._arraysize

    @arraysize.setter
    def arraysize(self, value):
        self._arraysize = int(value)

    @property
    def connection(self):
        return self._connection

    @property
    def errorhandler(self):
        return self._errorhandler

    @errorhandler.setter
    def errorhandler(self, value):
        logger.debug("setting errorhandler: %s", value)
        if value is None:
            raise ProgrammingError("Invalid errorhandler is specified")
        self._errorhandler = value

    @property
    def is_file_transfer(self):
        """Whether the command is PUT or GET."""
        return hasattr(self, "_is_file_transfer") and self._is_file_transfer

    def callproc(self, procname, args=()):
        """Not supported."""
        Error.errorhandler_wrapper(
            self.connection,
            self,
            NotSupportedError,
            {
                "msg": "callproc is not supported.",
                "errno": ER_UNSUPPORTED_METHOD,
                "sqlstate": SQLSTATE_FEATURE_NOT_SUPPORTED,
            },
        )

    def close(self) -> Optional[bool]:
        """Closes the cursor object.

        Returns whether the cursor was closed during this call.
        """
        try:
            if self.is_closed():
                return False

            with self._lock_canceling:
                self.reset()
                self._connection = None
                del self.messages[:]
                return True
        except Exception:
            pass

    def is_closed(self):
        return self._connection is None or self._connection.is_closed()

    def _execute_helper(
        self,
        query: str,
        timeout: int = 0,
        statement_params: Optional[Dict[str, str]] = None,
        binding_params: Union[Tuple, Dict[str, Dict[str, str]]] = None,
        binding_stage: Optional[str] = None,
        is_internal: bool = False,
        describe_only: bool = False,
        _no_results: bool = False,
        _is_put_get=None,
    ):
        del self.messages[:]

        if statement_params is not None and not isinstance(statement_params, dict):
            Error.errorhandler_wrapper(
                self.connection,
                self,
                ProgrammingError,
                {
                    "msg": "The data type of statement params is invalid. It must be dict.",
                    "errno": ER_INVALID_VALUE,
                },
            )

        # check if current installation include arrow extension or not,
        # if not, we set statement level query result format to be JSON
        if not CAN_USE_ARROW_RESULT_FORMAT:
            logger.debug("Cannot use arrow result format, fallback to json format")
            if statement_params is None:
                statement_params = {
                    PARAMETER_PYTHON_CONNECTOR_QUERY_RESULT_FORMAT: "JSON"
                }
            else:
                result_format_val = statement_params.get(
                    PARAMETER_PYTHON_CONNECTOR_QUERY_RESULT_FORMAT
                )
                if str(result_format_val).upper() == "ARROW":
                    self.check_can_use_arrow_resultset()
                elif result_format_val is None:
                    statement_params[
                        PARAMETER_PYTHON_CONNECTOR_QUERY_RESULT_FORMAT
                    ] = "JSON"

        self._sequence_counter = self._connection._next_sequence_counter()
        self._request_id = uuid.uuid4()

        logger.debug(f"Request id: {self._request_id}")

        if logger.getEffectiveLevel() <= logging.DEBUG:
            logger.debug("running query [%s]", self._format_query_for_log(query))
        if _is_put_get is not None:
            # if told the query is PUT or GET, use the information
            self._is_file_transfer = _is_put_get
        else:
            # or detect it.
            self._is_file_transfer = self.PUT_SQL_RE.match(
                query
            ) or self.GET_SQL_RE.match(query)
        logger.debug("is_file_transfer: %s", self._is_file_transfer is not None)

        real_timeout = (
            timeout if timeout and timeout > 0 else self._connection.network_timeout
        )

        if real_timeout is not None:
            self._timebomb = Timer(real_timeout, self.__cancel_query, [query])
            self._timebomb.start()
            logger.debug("started timebomb in %ss", real_timeout)
        else:
            self._timebomb = None

        original_sigint = signal.getsignal(signal.SIGINT)

        def interrupt_handler(*_):  # pragma: no cover
            try:
                signal.signal(signal.SIGINT, exit_handler)
            except (ValueError, TypeError):
                # ignore failures
                pass
            try:
                if self._timebomb is not None:
                    self._timebomb.cancel()
                    logger.debug("cancelled timebomb in finally")
                    self._timebomb = None
                self.__cancel_query(query)
            finally:
                if original_sigint:
                    try:
                        signal.signal(signal.SIGINT, original_sigint)
                    except (ValueError, TypeError):
                        # ignore failures
                        pass
            raise KeyboardInterrupt

        try:
            if not original_sigint == exit_handler:
                signal.signal(signal.SIGINT, interrupt_handler)
        except ValueError:  # pragma: no cover
            logger.debug(
                "Failed to set SIGINT handler. " "Not in main thread. Ignored..."
            )
        ret = {"data": {}}
        try:
            ret = self._connection.cmd_query(
                query,
                self._sequence_counter,
                self._request_id,
                binding_params=binding_params,
                binding_stage=binding_stage,
                is_file_transfer=bool(self._is_file_transfer),
                statement_params=statement_params,
                is_internal=is_internal,
                describe_only=describe_only,
                _no_results=_no_results,
            )
        finally:
            try:
                if original_sigint:
                    signal.signal(signal.SIGINT, original_sigint)
            except (ValueError, TypeError):  # pragma: no cover
                logger.debug(
                    "Failed to reset SIGINT handler. Not in main " "thread. Ignored..."
                )
            except Exception:
                self.connection.incident.report_incident()
                raise
            if self._timebomb is not None:
                self._timebomb.cancel()
                logger.debug("cancelled timebomb in finally")

        if "data" in ret and "parameters" in ret["data"]:
            parameters = ret["data"]["parameters"]
            # Set session parameters for cursor object
            for kv in parameters:
                if "TIMESTAMP_OUTPUT_FORMAT" in kv["name"]:
                    self._timestamp_output_format = kv["value"]
                if "TIMESTAMP_NTZ_OUTPUT_FORMAT" in kv["name"]:
                    self._timestamp_ntz_output_format = kv["value"]
                if "TIMESTAMP_LTZ_OUTPUT_FORMAT" in kv["name"]:
                    self._timestamp_ltz_output_format = kv["value"]
                if "TIMESTAMP_TZ_OUTPUT_FORMAT" in kv["name"]:
                    self._timestamp_tz_output_format = kv["value"]
                if "DATE_OUTPUT_FORMAT" in kv["name"]:
                    self._date_output_format = kv["value"]
                if "TIME_OUTPUT_FORMAT" in kv["name"]:
                    self._time_output_format = kv["value"]
                if "TIMEZONE" in kv["name"]:
                    self._timezone = kv["value"]
                if "BINARY_OUTPUT_FORMAT" in kv["name"]:
                    self._binary_output_format = kv["value"]
            # Set session parameters for connection object
            self._connection._update_parameters(
                {p["name"]: p["value"] for p in parameters}
            )

        self._sequence_counter = -1
        return ret

    def execute(
        self,
        command: str,
        params: Optional[Sequence] = None,
        _bind_stage: Optional[str] = None,
        timeout: Optional[int] = None,
        _exec_async: bool = False,
        _do_reset: bool = True,
        _put_callback: "SnowflakeProgressPercentage" = None,
        _put_azure_callback: "SnowflakeProgressPercentage" = None,
        _put_callback_output_stream: IO[str] = sys.stdout,
        _get_callback: "SnowflakeProgressPercentage" = None,
        _get_azure_callback: "SnowflakeProgressPercentage" = None,
        _get_callback_output_stream: IO[str] = sys.stdout,
        _show_progress_bar: bool = True,
        _statement_params: Optional[Dict[str, str]] = None,
        _is_internal: bool = False,
        _describe_only: bool = False,
        _no_results: bool = False,
        _use_ijson: bool = False,
        _is_put_get: Optional[bool] = None,
        _raise_put_get_error: bool = True,
        _force_put_overwrite: bool = False,
        file_stream: Optional[IO[bytes]] = None,
    ) -> Optional[Union["SnowflakeCursor", None]]:
        """Executes a command/query.

        Args:
            command: The SQL command to be executed.
            params: Parameters to be bound into the SQL statement.
            _bind_stage: Path in temporary stage where binding parameters are uploaded as CSV files.
            timeout: Number of seconds after which to abort the query.
            _exec_async: Whether to execute this query asynchronously.
            _do_reset: Whether or not the result set needs to be reset before executing query.
            _put_callback: Function to which GET command should call back to.
            _put_azure_callback: Function to which an Azure GET command should call back to.
            _put_callback_output_stream: The output stream a PUT command's callback should report on.
            _get_callback: Function to which GET command should call back to.
            _get_azure_callback: Function to which an Azure GET command should call back to.
            _get_callback_output_stream: The output stream a GET command's callback should report on.
            _show_progress_bar: Whether or not to show progress bar.
            _statement_params: Extra information that should be sent to Snowflake with query.
            _is_internal: This flag indicates whether the query is issued internally by the connector.
            _describe_only: If true, the query will not be executed but return the schema/description of this query.
            _no_results: This flag tells the back-end to not return the result, just fire the query and return the
                response returned by Snowflake's server.
            _use_ijson: This flag doesn't do anything as ijson support has ended.
            _is_put_get: Force decision of this SQL query being a PUT, or GET command. This is detected otherwise.
            _raise_put_get_error: Whether to raise PUT and GET errors.
            _force_put_overwrite: If the SQL query is a PUT, then this flag can force overwriting of an already
                existing file on stage.
            file_stream: File-like object to be uploaded with PUT

        Returns:
            The cursor itself, or None if some error happened, or the response returned
            by Snowflake if the _no_results flag is on.
        """
        if _exec_async:
            _no_results = True
        logger.debug("executing SQL/command")
        if self.is_closed():
            Error.errorhandler_wrapper(
                self.connection,
                self,
                DatabaseError,
                {"msg": "Cursor is closed in execute.", "errno": ER_CURSOR_IS_CLOSED},
            )

        if _do_reset:
            self.reset()
        command = command.strip(" \t\n\r") if command else None
        if not command:
            logger.warning("execute: no query is given to execute")
            return

        kwargs = {
            "timeout": timeout,
            "statement_params": _statement_params,
            "is_internal": _is_internal,
            "describe_only": _describe_only,
            "_no_results": _no_results,
            "_is_put_get": _is_put_get,
        }

        try:
            if self._connection.is_pyformat:
                # pyformat/format paramstyle
                # client side binding
                processed_params = self._connection._process_params_pyformat(
                    params, self
                )
                if logger.getEffectiveLevel() <= logging.DEBUG:
                    logger.debug(
                        f"binding: [{self._format_query_for_log(command)}] "
                        f"with input=[{params}], "
                        f"processed=[{processed_params}]",
                    )
                if len(processed_params) > 0:
                    query = command % processed_params
                else:
                    query = command
            else:
                # qmark and numeric paramstyle
                query = command
                if _bind_stage:
                    kwargs["binding_stage"] = _bind_stage
                else:
                    kwargs["binding_params"] = self._connection._process_params_qmarks(
                        params, self
                    )
        # Skip reporting Key, Value and Type errors
        except Exception as exc:  # pragma: no cover
            if not isinstance(exc, INCIDENT_BLACKLIST):
                self.connection.incident.report_incident()
            raise

        m = DESC_TABLE_RE.match(query)
        if m:
            query1 = "describe table {}".format(m.group(1))
            if logger.getEffectiveLevel() <= logging.WARNING:
                logger.info(
                    "query was rewritten: org=%s, new=%s",
                    " ".join(line.strip() for line in query.split("\n")),
                    query1,
                )
            query = query1

        if logger.getEffectiveLevel() <= logging.INFO:
            logger.info("query: [%s]", self._format_query_for_log(query))
        ret = self._execute_helper(query, **kwargs)
        self._sfqid = (
            ret["data"]["queryId"]
            if "data" in ret and "queryId" in ret["data"]
            else None
        )
        self._sqlstate = (
            ret["data"]["sqlState"]
            if "data" in ret and "sqlState" in ret["data"]
            else None
        )
        self._first_chunk_time = get_time_millis()

        # if server gives a send time, log the time it took to arrive
        if "data" in ret and "sendResultTime" in ret["data"]:
            time_consume_first_result = (
                self._first_chunk_time - ret["data"]["sendResultTime"]
            )
            self._log_telemetry_job_data(
                TelemetryField.TIME_CONSUME_FIRST_RESULT, time_consume_first_result
            )
        logger.debug("sfqid: %s", self.sfqid)

        logger.info("query execution done")
        if ret["success"]:
            logger.debug("SUCCESS")
            data = ret["data"]

            logger.debug("PUT OR GET: %s", self.is_file_transfer)
            if self.is_file_transfer:
                # Decide whether to use the old, or new code path
                if self._connection._use_new_put_get:
                    sf_file_transfer_agent_class = SnowflakeFileTransferAgent
                else:
                    sf_file_transfer_agent_class = SnowflakeFileTransferAgentSdk

                sf_file_transfer_agent = sf_file_transfer_agent_class(
                    self,
                    query,
                    ret,
                    put_callback=_put_callback,
                    put_azure_callback=_put_azure_callback,
                    put_callback_output_stream=_put_callback_output_stream,
                    get_callback=_get_callback,
                    get_azure_callback=_get_azure_callback,
                    get_callback_output_stream=_get_callback_output_stream,
                    show_progress_bar=_show_progress_bar,
                    raise_put_get_error=_raise_put_get_error,
                    force_put_overwrite=_force_put_overwrite
                    or data.get("overwrite", False),
                    source_from_stream=file_stream,
                    multipart_threshold=data.get("threshold"),
                    use_s3_regional_url=self._connection.enable_stage_s3_privatelink_for_us_east_1,
                )
                sf_file_transfer_agent.execute()
                data = sf_file_transfer_agent.result()
                self._total_rowcount = len(data["rowset"]) if "rowset" in data else -1
            m = self.ALTER_SESSION_RE.match(query)
            if m:
                # session parameters
                param = m.group(1).upper()
                value = m.group(2)
                self._connection.converter.set_parameter(param, value)

            if _exec_async:
                self.connection._async_sfqids.add(self._sfqid)
            if _no_results:
                self._total_rowcount = (
                    ret["data"]["total"]
                    if "data" in ret and "total" in ret["data"]
                    else -1
                )
                return data
            self._init_result_and_meta(data)
        else:
            self._total_rowcount = (
                ret["data"]["total"] if "data" in ret and "total" in ret["data"] else -1
            )
            logger.debug(ret)
            err = ret["message"]
            code = ret.get("code", -1)
            if "data" in ret:
                err += ret["data"].get("errorMessage", "")
            errvalue = {
                "msg": err,
                "errno": int(code),
                "sqlstate": self._sqlstate,
                "sfqid": self._sfqid,
            }
            Error.errorhandler_wrapper(
                self.connection, self, ProgrammingError, errvalue
            )
        return self

    def execute_async(self, *args, **kwargs):
        """Convenience function to execute a query without waiting for results (asynchronously).

        This function takes the same arguments as execute, please refer to that function
        for documentation.
        """
        kwargs["_exec_async"] = True
        return self.execute(*args, **kwargs)

    def describe(self, *args, **kwargs) -> List[ResultMetadata]:
        """Obtain the schema of the result without executing the query.

        This function takes the same arguments as execute, please refer to that function
        for documentation.

        Returns:
            The schema of the result.
        """
        kwargs["_describe_only"] = kwargs["_is_internal"] = True
        self.execute(*args, **kwargs)
        return self._description

    def _format_query_for_log(self, query):
        return self._connection._format_query_for_log(query)

    def _is_dml(self, data):
        return (
            "statementTypeId" in data
            and int(data["statementTypeId"]) in STATEMENT_TYPE_ID_DML_SET
        )

    def _init_result_and_meta(self, data):
        is_dml = self._is_dml(data)
        self._query_result_format = data.get("queryResultFormat", "json")
        logger.debug("Query result format: %s", self._query_result_format)

        if self._total_rowcount == -1 and not is_dml and data.get("total") is not None:
            self._total_rowcount = data["total"]

        self._description: List[ResultMetadata] = [
            ResultMetadata.from_column(col) for col in data["rowtype"]
        ]

        result_chunks = create_batches_from_response(
            self, self._query_result_format, data, self._description
        )

        self._result_set = ResultSet(
            self,
            result_chunks,
        )
        self._rownumber = -1
        self._result_state = ResultState.VALID

        # don't update the row count when the result is returned from `describe` method
        if is_dml and "rowset" in data and len(data["rowset"]) > 0:
            updated_rows = 0
            for idx, desc in enumerate(self._description):
                if (
                    desc[0]
                    in (
                        "number of rows updated",
                        "number of multi-joined rows updated",
                        "number of rows deleted",
                    )
                    or desc[0].startswith("number of rows inserted")
                ):
                    updated_rows += int(data["rowset"][0][idx])
            if self._total_rowcount == -1:
                self._total_rowcount = updated_rows
            else:
                self._total_rowcount += updated_rows

    def check_can_use_arrow_resultset(self):
        global CAN_USE_ARROW_RESULT_FORMAT

        if not CAN_USE_ARROW_RESULT_FORMAT:
            if self._connection.application == "SnowSQL":
                msg = "Currently SnowSQL doesn't support the result set in Apache Arrow format."
                errno = ER_NO_PYARROW_SNOWSQL
            else:
                msg = "The result set in Apache Arrow format is not supported for the platform."
                errno = ER_NO_ARROW_RESULT

            Error.errorhandler_wrapper(
                self.connection,
                self,
                ProgrammingError,
                {
                    "msg": msg,
                    "errno": errno,
                },
            )

    def check_can_use_pandas(self):
        if not installed_pandas:
            msg = (
                "Optional dependency: 'pyarrow' is not installed, please see the following link for install "
                "instructions: https://docs.snowflake.com/en/user-guide/python-connector-pandas.html#installation"
            )
            errno = ER_NO_PYARROW

            Error.errorhandler_wrapper(
                self.connection,
                self,
                ProgrammingError,
                {
                    "msg": msg,
                    "errno": errno,
                },
            )

    def query_result(self, qid):
        url = "/queries/{qid}/result".format(qid=qid)
        ret = self._connection.rest.request(url=url, method="get")
        self._sfqid = (
            ret["data"]["queryId"]
            if "data" in ret and "queryId" in ret["data"]
            else None
        )
        self._sqlstate = (
            ret["data"]["sqlState"]
            if "data" in ret and "sqlState" in ret["data"]
            else None
        )
        logger.debug("sfqid=%s", self._sfqid)

        if ret.get("success"):
            data = ret.get("data")
            self._init_result_and_meta(data)
        else:
            logger.info("failed")
            logger.debug(ret)
            err = ret["message"]
            code = ret.get("code", -1)
            if "data" in ret:
                err += ret["data"].get("errorMessage", "")
            errvalue = {
                "msg": err,
                "errno": int(code),
                "sqlstate": self._sqlstate,
                "sfqid": self._sfqid,
            }
            Error.errorhandler_wrapper(
                self.connection, self, ProgrammingError, errvalue
            )
        return self

    def fetch_arrow_batches(self) -> Iterator[Table]:
        self.check_can_use_arrow_resultset()
        if self._query_result_format != "arrow":
            raise NotSupportedError
        self._log_telemetry_job_data(
            TelemetryField.ARROW_FETCH_BATCHES, TelemetryData.TRUE
        )
        return self._result_set._fetch_arrow_batches()

    def fetch_arrow_all(self) -> Optional[Table]:
        self.check_can_use_arrow_resultset()
        if self._query_result_format != "arrow":
            raise NotSupportedError
        self._log_telemetry_job_data(TelemetryField.ARROW_FETCH_ALL, TelemetryData.TRUE)
        return self._result_set._fetch_arrow_all()

    def fetch_pandas_batches(self, **kwargs) -> Iterator["pandas.DataFrame"]:
        """Fetches a single Arrow Table."""
        self.check_can_use_pandas()
        if self._prefetch_hook is not None:
            self._prefetch_hook()
            raise NotSupportedError
        self._log_telemetry_job_data(
            TelemetryField.PANDAS_FETCH_BATCHES, TelemetryData.TRUE
        )
        return self._result_set._fetch_pandas_batches(**kwargs)

    def fetch_pandas_all(self, **kwargs) -> "pandas.DataFrame":
        """Fetch Pandas dataframes in batches, where 'batch' refers to Snowflake Chunk."""
        self.check_can_use_pandas()
        if self._prefetch_hook is not None:
            self._prefetch_hook()
        if self._query_result_format != "arrow":
            raise NotSupportedError
        self._log_telemetry_job_data(
            TelemetryField.PANDAS_FETCH_ALL, TelemetryData.TRUE
        )
        return self._result_set._fetch_pandas_all(**kwargs)

    def abort_query(self, qid):
        url = "/queries/{qid}/abort-request".format(qid=qid)
        ret = self._connection.rest.request(url=url, method="post")
        return ret.get("success")

    def executemany(
        self,
        command: str,
        seqparams: Union[Sequence[Any], Dict[str, Any]],
    ) -> "SnowflakeCursor":
        """Executes a command/query with the given set of parameters sequentially."""
        logger.debug("executing many SQLs/commands")
        command = command.strip(" \t\n\r") if command else None

        if len(seqparams) == 0:
            Error.errorhandler_wrapper(
                self.connection,
                self,
                InterfaceError,
                {
                    "msg": f"No parameters are specified for the command: {command}",
                    "errno": ER_INVALID_VALUE,
                },
            )
            return self

        if self.INSERT_SQL_RE.match(command):
            if self._connection.is_pyformat:
                logger.debug("rewriting INSERT query")
                command_wo_comments = re.sub(self.COMMENT_SQL_RE, "", command)
                m = self.INSERT_SQL_VALUES_RE.match(command_wo_comments)
                if not m:
                    Error.errorhandler_wrapper(
                        self.connection,
                        self,
                        InterfaceError,
                        {
                            "msg": "Failed to rewrite multi-row insert",
                            "errno": ER_FAILED_TO_REWRITE_MULTI_ROW_INSERT,
                        },
                    )

                fmt = m.group(1)
                values = []
                for param in seqparams:
                    logger.debug(f"parameter: {param}")
                    values.append(
                        fmt % self._connection._process_params_pyformat(param, self)
                    )
                command = command.replace(fmt, ",".join(values), 1)
                self.execute(command)
                return self
            else:
                logger.debug("bulk insert")
                # sanity check
                row_size = len(seqparams[0])
                for row in seqparams:
                    if len(row) != row_size:
                        error_value = {
                            "msg": f"Bulk data size don't match. expected: {row_size}, "
                            f"got: {len(row)}, command: {command}",
                            "errno": ER_INVALID_VALUE,
                        }
                        Error.errorhandler_wrapper(
                            self.connection, self, InterfaceError, error_value
                        )
                        return self
                bind_size = len(seqparams) * row_size
                bind_stage = None
                if (
                    bind_size
                    > self.connection._session_parameters[
                        "CLIENT_STAGE_ARRAY_BINDING_THRESHOLD"
                    ]
                    > 0
                ):
                    # bind stage optimization
                    try:
                        rows = self.connection._write_params_to_byte_rows(seqparams)
                        bind_uploader = BindUploadAgent(self, rows)
                        bind_uploader.upload()
                        bind_stage = bind_uploader.stage_path
                    except BindUploadError:
                        logger.debug(
                            "Failed to upload binds to stage, sending binds to "
                            "Snowflake instead."
                        )
                    except Exception as exc:
                        if not isinstance(exc, INCIDENT_BLACKLIST):
                            self.connection.incident.report_incident()
                        raise
                binding_param = (
                    None if bind_stage else list(map(list, zip(*seqparams)))
                )  # transpose
                self.execute(command, params=binding_param, _bind_stage=bind_stage)
                return self

        self.reset()
        for param in seqparams:
            self.execute(command, param, _do_reset=False)
        return self

    def _result_iterator(
        self,
    ) -> Union[Generator[Dict, None, None], Generator[Tuple, None, None]]:
        """Yields the elements from _result and raises an exception when appropriate."""
        try:
            for _next in self._result:
                if isinstance(_next, Exception):
                    Error.errorhandler_wrapper_from_ready_exception(
                        self._connection,
                        self,
                        _next,
                    )
                self._rownumber += 1
                yield _next
        except TypeError as err:
            if self._result_state == ResultState.DEFAULT:
                raise err
            else:
                yield None

    def fetchone(self) -> Optional[Union[Dict, Tuple]]:
        """Fetches one row."""
        if self._prefetch_hook is not None:
            self._prefetch_hook()
        if self._result is None and self._result_set is not None:
            self._result = iter(self._result_set)
            self._result_state = ResultState.VALID
        try:
            return next(self._result_iterator())
        except StopIteration:
            return None

    def fetchmany(self, size=None):
        """Fetches the number of specified rows."""
        if size is None:
            size = self.arraysize

        if size < 0:
            errorvalue = {
                "msg": (
                    "The number of rows is not zero or " "positive number: {}"
                ).format(size),
                "errno": ER_NOT_POSITIVE_SIZE,
            }
            Error.errorhandler_wrapper(
                self.connection, self, ProgrammingError, errorvalue
            )
        ret = []
        while size > 0:
            row = self.fetchone()
            if row is None:
                break
            ret.append(row)
            if size is not None:
                size -= 1

        return ret

    def fetchall(self) -> Union[List[Tuple], List[Dict]]:
        """Fetches all of the results."""
        ret = []
        while True:
            row = self.fetchone()
            if row is None:
                break
            ret.append(row)
        return ret

    def nextset(self):
        """Not supported."""
        logger.debug("nop")
        return None

    def setinputsizes(self, _):
        """Not supported."""
        logger.debug("nop")

    def setoutputsize(self, _, column=None):
        """Not supported."""
        del column
        logger.debug("nop")

    def scroll(self, value, mode="relative"):
        Error.errorhandler_wrapper(
            self.connection,
            self,
            NotSupportedError,
            {
                "msg": "scroll is not supported.",
                "errno": ER_UNSUPPORTED_METHOD,
                "sqlstate": SQLSTATE_FEATURE_NOT_SUPPORTED,
            },
        )

    def reset(self):
        """Resets the result set."""
        self._total_rowcount = -1  # reset the rowcount
        if self._result_state != ResultState.DEFAULT:
            self._result_state = ResultState.RESET
        if self._result is not None:
            self._result = None
        if self._inner_cursor is not None:
            self._inner_cursor.reset()
            self._result = None
            self._inner_cursor = None
        self._prefetch_hook = None
        if not self.connection._reuse_results:
            self._result_set = None

    def __iter__(self) -> Union[Iterator[Dict], Iterator[Tuple]]:
        """Iteration over the result set."""
        # set _result if _result_set is not None
        if self._result is None and self._result_set is not None:
            self._result = iter(self._result_set)
            self._result_state = ResultState.VALID
        return self._result_iterator()

    def __cancel_query(self, query):
        if self._sequence_counter >= 0 and not self.is_closed():
            logger.debug("canceled. %s, request_id: %s", query, self._request_id)
            with self._lock_canceling:
                self._connection._cancel_query(query, self._request_id)

    def _log_telemetry_job_data(self, telemetry_field, value):
        """Builds an instance of TelemetryData with the given field and logs it."""
        obj = {
            "type": telemetry_field,
            "query_id": self._sfqid,
            "value": int(value),
        }
        ts = get_time_millis()
        try:
            self._connection._log_telemetry(TelemetryData(obj, ts))
        except AttributeError:
            logger.warning(
                "Cursor failed to log to telemetry. Connection object may be None.",
                exc_info=True,
            )

    def __enter__(self):
        """Context manager."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager with commit or rollback."""
        self.close()

    def get_results_from_sfqid(self, sfqid: str):
        """Gets the results from previously ran query."""

        def wait_until_ready():
            """Makes sure query has finished executing and once it has retrieves results."""
            no_data_counter = 0
            retry_pattern_pos = 0
            while True:
                status = self.connection.get_query_status(sfqid)
                if not self.connection.is_still_running(status):
                    break
                if status == QueryStatus.NO_DATA:  # pragma: no cover
                    no_data_counter += 1
                    if no_data_counter > ASYNC_NO_DATA_MAX_RETRY:
                        raise DatabaseError(
                            "Cannot retrieve data on the status of this query. No information returned "
                            "from server for query '{}'"
                        )
                time.sleep(
                    0.5 * ASYNC_RETRY_PATTERN[retry_pattern_pos]
                )  # Same wait as JDBC
                # If we can advance in ASYNC_RETRY_PATTERN then do so
                if retry_pattern_pos < (len(ASYNC_RETRY_PATTERN) - 1):
                    retry_pattern_pos += 1
            if status != QueryStatus.SUCCESS:
                raise DatabaseError(
                    "Status of query '{}' is {}, results are unavailable".format(
                        sfqid, status.name
                    )
                )
            self._inner_cursor.execute(
                "select * from table(result_scan('{}'))".format(sfqid)
            )
            self._result = self._inner_cursor._result
            self._query_result_format = self._inner_cursor._query_result_format
            self._total_rowcount = self._inner_cursor._total_rowcount
            self._description = self._inner_cursor._description
            self._result_set = self._inner_cursor._result_set
            self._result_state = ResultState.VALID
            self._rownumber = 0
            # Unset this function, so that we don't block anymore
            self._prefetch_hook = None

        self.connection.get_query_status_throw_if_error(
            sfqid
        )  # Trigger an exception if query failed
        klass = self.__class__
        self._inner_cursor = klass(self.connection)
        self._sfqid = sfqid
        self._prefetch_hook = wait_until_ready

    def get_result_batches(self) -> Optional[List["ResultBatch"]]:
        """Get the previously executed query's ``ResultBatch`` s if available.

        If they are unavailable, in case nothing has been executed yet None will
        be returned.

        For a detailed description of ``ResultBatch`` s please see the docstring of:
        ``snowflake.connector.result_batches.ResultBatch``
        """
        if self._result_set is None:
            return None
        self._log_telemetry_job_data(
            TelemetryField.GET_PARTITIONS_USED, TelemetryData.TRUE
        )
        return self._result_set.batches


class DictCursor(SnowflakeCursor):
    """Cursor returning results in a dictionary."""

    def __init__(self, connection):
        super().__init__(
            connection,
            use_dict_result=True,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2021 Snowflake Computing Inc. All right reserved.
#

import json
import os
from collections import namedtuple
from logging import getLogger
from typing import TYPE_CHECKING, Any, Dict, Optional

from .compat import quote
from .constants import HTTP_HEADER_CONTENT_ENCODING, FileHeader, ResultStatus
from .encryption_util import EncryptionMetadata
from .vendored import requests

if TYPE_CHECKING:  # pragma: no cover
    from .file_transfer_agent_sdk import SnowflakeFileMeta

logger = getLogger(__name__)

GCS_METADATA_PREFIX = "x-goog-meta-"
GCS_METADATA_SFC_DIGEST = GCS_METADATA_PREFIX + "sfc-digest"
GCS_METADATA_MATDESC_KEY = GCS_METADATA_PREFIX + "matdesc"
GCS_METADATA_ENCRYPTIONDATAPROP = GCS_METADATA_PREFIX + "encryptiondata"
GCS_FILE_HEADER_DIGEST = "gcs-file-header-digest"
GCS_FILE_HEADER_CONTENT_LENGTH = "gcs-file-header-content-length"
GCS_FILE_HEADER_ENCRYPTION_METADATA = "gcs-file-header-encryption-metadata"
CONTENT_CHUNK_SIZE = 10 * 1024

"""
Gcs Location: Gcs bucket name + path
"""
GcsLocation = namedtuple(
    "GcsLocation", ["bucket_name", "path"]  # Gcs bucket name  # Gcs path name
)


class SnowflakeGCSUtil:
    """GCS Utility class."""

    @staticmethod
    def create_client(
        stage_info: Dict[str, Any],
        use_accelerate_endpoint: bool = False,
        use_s3_regional_url: bool = False,
        s3_connection_pool_size: int = 1,
    ) -> Optional[str]:
        """Creates a client object with given stage credentials.

        Args:
            stage_info: Access credentials and info of a stage.
            use_accelerate_endpoint: Whether to use an accelerated endpoint? This is not applicable to GCS.

        Returns:
            The client to communicate with GCS.
        """
        stage_credentials = stage_info["creds"]
        security_token = stage_credentials.get("GCS_ACCESS_TOKEN")

        if security_token:
            logger.debug(f"len(GCS_ACCESS_TOKEN): {len(security_token)}")
            logger.debug("Access token is saved as client for renew")
            client = security_token

        else:
            logger.debug(
                "No access token received from GS, constructing anonymous client"
            )
            client = None

        return client

    @staticmethod
    def upload_file(
        data_file: str,
        meta: "SnowflakeFileMeta",
        encryption_metadata: Any,
        max_concurrency: int,
        multipart_threshold: int,
    ):
        """Uploads the local file to GCS's blob storage.

        Args:
            data_file: File path on local system.
            meta: The File meta object (contains credentials and remote location).
            encryption_metadata: Encryption metadata to be set on object.
            max_concurrency: The maximum number of threads to used to upload. Not applicable to GCS.
            multipart_threshold: The number of bytes after which size a file should be uploaded concurrently in chunks.
                Not applicable to GCS.

        Raises:
            HTTPError if some http errors occurred.

        Returns:
            None.
        """
        upload_url = meta.presigned_url
        access_token: Optional[str] = None

        if not upload_url:
            upload_url = SnowflakeGCSUtil.generate_file_url(
                meta.client_meta.stage_info["location"], meta.dst_file_name.lstrip("/")
            )
            access_token = meta.client_meta.cloud_client

        content_encoding = ""
        if meta.dst_compression_type is not None:
            content_encoding = meta.dst_compression_type.name.lower()

        # We set the contentEncoding to blank for GZIP files. We don't
        # want GCS to think our gzip files are gzips because it makes
        # them download uncompressed, and none of the other providers do
        # that. There's essentially no way for us to prevent that
        # behavior. Bad Google.
        if content_encoding and content_encoding == "gzip":
            content_encoding = ""

        gcs_headers = {
            HTTP_HEADER_CONTENT_ENCODING: content_encoding,
            GCS_METADATA_SFC_DIGEST: meta.sha256_digest,
        }
        if access_token:
            gcs_headers.update({"Authorization": f"Bearer {access_token}"})

        if encryption_metadata:
            gcs_headers.update(
                {
                    GCS_METADATA_ENCRYPTIONDATAPROP: json.dumps(
                        {
                            "EncryptionMode": "FullBlob",
                            "WrappedContentKey": {
                                "KeyId": "symmKey1",
                                "EncryptedKey": encryption_metadata.key,
                                "Algorithm": "AES_CBC_256",
                            },
                            "EncryptionAgent": {
                                "Protocol": "1.0",
                                "EncryptionAlgorithm": "AES_CBC_256",
                            },
                            "ContentEncryptionIV": encryption_metadata.iv,
                            "KeyWrappingMetadata": {"EncryptionLibrary": "Java 5.3.0"},
                        }
                    ),
                    GCS_METADATA_MATDESC_KEY: encryption_metadata.matdesc,
                }
            )

        try:
            upload_src = None
            if meta.src_stream is None:
                upload_src = open(data_file, "rb")
            else:
                upload_src = meta.real_src_stream or meta.src_stream

            response = requests.put(
                url=upload_url, data=upload_src, headers=gcs_headers
            )
            response.raise_for_status()

        except requests.exceptions.HTTPError as errh:
            logger.debug("GCS file upload Http Error: %s", errh)
            # Presigned urls can be generated for any xml-api operation
            # offered by GCS. Hence the error codes expected are similar
            # to xml api.
            # https://cloud.google.com/storage/docs/xml-api/reference-status

            # According to the above resource, GCS recommends retrying
            # for the following error codes.
            if errh.response.status_code in [403, 408, 429, 500, 503]:
                meta.last_error = errh
                meta.result_status = ResultStatus.NEED_RETRY
                return
            elif (
                (not access_token)
                and errh.response.status_code == 400
                and (
                    meta.last_error is None
                    or meta.last_error.response.status_code != 400
                )
            ):
                # Only attempt to renew urls if this isn't the second time this happens
                meta.last_error = errh
                meta.result_status = ResultStatus.RENEW_PRESIGNED_URL
                return
            elif access_token and SnowflakeGCSUtil.is_token_expired(errh.response):
                meta.last_error = errh
                meta.result_status = ResultStatus.RENEW_TOKEN
                return
            # raise anything else
            raise errh
        except requests.exceptions.Timeout as errt:
            logger.debug(f"GCS file upload Timeout Error: {errt}")
            meta.last_error = errt
            meta.result_status = ResultStatus.NEED_RETRY
            return
        finally:
            if meta.src_stream is None:
                upload_src.close()

        if meta.put_callback is not None:
            meta.put_callback(
                data_file,
                meta.src_file_size,
                output_stream=meta.put_callback_output_stream,
                show_progress_bar=meta.show_progress_bar,
            )(meta.src_file_size)

        logger.debug("DONE putting a file")
        meta.dst_file_size = meta.upload_size
        meta.result_status = ResultStatus.UPLOADED

        meta.gcs_file_header_digest = gcs_headers[GCS_METADATA_SFC_DIGEST]
        meta.gcs_file_header_content_length = meta.upload_size
        meta.gcs_file_header_encryption_metadata = json.loads(
            gcs_headers.get(GCS_METADATA_ENCRYPTIONDATAPROP, "null")
        )

    @staticmethod
    def _native_download_file(
        meta: "SnowflakeFileMeta", full_dst_file_name: str, max_concurrency: int
    ):
        """Downloads the remote object to local file.

        Args:
            meta: File meta object (contains credentials and remote location).
            full_dst_file_name: Local path of the file to download to.
            max_concurrency: Not applicable to GCS.

        Raises:
            HTTPError if some http errors occurred.

        Returns:
            None, if downloading was successful.
        """
        download_url = meta.presigned_url
        gcs_headers = None
        access_token: Optional[str] = None

        if not download_url:
            download_url = SnowflakeGCSUtil.generate_file_url(
                meta.client_meta.stage_info["location"], meta.src_file_name.lstrip("/")
            )
            access_token = meta.client_meta.cloud_client
            gcs_headers = {"Authorization": f"Bearer {access_token}"}

        try:
            response = requests.get(download_url, headers=gcs_headers, stream=True)
            response.raise_for_status()

            with open(full_dst_file_name, "wb") as fd:
                for chunk in response.raw.stream(
                    CONTENT_CHUNK_SIZE, decode_content=False
                ):
                    fd.write(chunk)
        except requests.exceptions.HTTPError as errh:
            logger.debug(f"GCS file download Http Error: {errh}")
            # Presigned urls can be generated for any xml-api operation
            # offered by GCS. Hence the error codes expected are similar
            # to xml api.
            # https://cloud.google.com/storage/docs/xml-api/reference-status

            # According to the above resource, GCS recommends retrying
            # for the following error codes.
            if errh.response.status_code in [403, 408, 429, 500, 503]:
                meta.last_error = errh
                meta.result_status = ResultStatus.NEED_RETRY
                return
            elif (
                (not access_token)
                and errh.response.status_code == 400
                and (
                    meta.last_error is None
                    or meta.last_error.errh.response.status_code != 400
                )
            ):
                # Only attempt to renew urls if this isn't the second time this happens
                meta.last_error = errh
                meta.result_status = ResultStatus.RENEW_PRESIGNED_URL
                return
            elif access_token and SnowflakeGCSUtil.is_token_expired(errh.response):
                meta.last_error = errh
                meta.result_status = ResultStatus.RENEW_TOKEN
                return
            # raise anything else
            raise errh
        except requests.exceptions.Timeout as errt:
            logger.debug(f"GCS file download Timeout Error: {errt}")
            meta.last_error = errt
            meta.result_status = ResultStatus.NEED_RETRY
            return

        encryption_metadata = None

        if response.headers.get(GCS_METADATA_ENCRYPTIONDATAPROP, None):
            encryptiondata = json.loads(
                response.headers[GCS_METADATA_ENCRYPTIONDATAPROP]
            )

            if encryptiondata:
                encryption_metadata = EncryptionMetadata(
                    key=encryptiondata["WrappedContentKey"]["EncryptedKey"],
                    iv=encryptiondata["ContentEncryptionIV"],
                    matdesc=response.headers[GCS_METADATA_MATDESC_KEY]
                    if GCS_METADATA_MATDESC_KEY in response.headers
                    else None,
                )

        # Sadly, we can only determine the src file size after we've
        # downloaded it, unlike the other cloud providers where the
        # metadata can be read beforehand.
        meta.src_file_size = os.path.getsize(full_dst_file_name)

        if meta.get_callback:
            meta.get_callback(
                meta.src_file_name,
                meta.src_file_size,
                output_stream=meta.get_callback_output_stream,
                show_progress_bar=meta.show_progress_bar,
            )(meta.src_file_size)

        logger.debug("DONE getting a file")
        meta.result_status = ResultStatus.DOWNLOADED

        meta.gcs_file_header_digest = response.headers.get(GCS_METADATA_SFC_DIGEST)
        meta.gcs_file_header_content_length = len(response.content)
        meta.gcs_file_header_encryption_metadata = encryption_metadata

    @staticmethod
    def get_file_header(
        meta: "SnowflakeFileMeta", filename: str
    ) -> Optional[FileHeader]:
        """Gets the remote file's metadata.

        Args:
            meta: Remote file's metadata info.
            filename: Not applicable to GCS.

        Returns:
            The file header, with expected properties populated or None, based on how the request goes with the
            storage provider.

        Notes:
            Sometimes this method is called to verify that the file has indeed been uploaded. In cases of presigned
            url, we have no way of verifying that, except with the http status code of 200 which we have already
            confirmed and set the meta.result_status = UPLOADED/DOWNLOADED.
        """
        if (
            meta.result_status == ResultStatus.UPLOADED
            or meta.result_status == ResultStatus.DOWNLOADED
        ):
            return FileHeader(
                digest=meta.gcs_file_header_digest,
                content_length=meta.gcs_file_header_content_length,
                encryption_metadata=meta.gcs_file_header_encryption_metadata,
            )
        else:
            if meta.presigned_url:
                meta.result_status = ResultStatus.NOT_FOUND_FILE
            else:
                url = SnowflakeGCSUtil.generate_file_url(
                    meta.client_meta.stage_info["location"], filename.lstrip("/")
                )
                access_token: str = meta.client_meta.cloud_client
                gcs_headers = {"Authorization": f"Bearer {access_token}"}
                try:
                    response = requests.head(url, headers=gcs_headers)
                    response.raise_for_status()

                    digest = response.headers.get(GCS_METADATA_SFC_DIGEST, None)
                    content_length = response.headers.get("content-length", None)
                    if response.headers.get(GCS_METADATA_ENCRYPTIONDATAPROP, None):
                        encryption_data = json.loads(
                            response.headers[GCS_METADATA_ENCRYPTIONDATAPROP]
                        )

                        if encryption_data:
                            encryption_metadata = EncryptionMetadata(
                                key=encryption_data["WrappedContentKey"][
                                    "EncryptedKey"
                                ],
                                iv=encryption_data["ContentEncryptionIV"],
                                matdesc=response.headers[GCS_METADATA_MATDESC_KEY]
                                if GCS_METADATA_MATDESC_KEY in response.headers
                                else None,
                            )
                    meta.result_status = ResultStatus.UPLOADED
                    return FileHeader(
                        digest=digest,
                        content_length=content_length,
                        encryption_metadata=encryption_metadata,
                    )
                except requests.exceptions.HTTPError as errh:
                    if errh.response.status_code in [403, 408, 429, 500, 503]:
                        meta.last_error = errh
                        meta.result_status = ResultStatus.NEED_RETRY
                        return
                    if errh.response.status_code == 404:
                        meta.result_status = ResultStatus.NOT_FOUND_FILE
                    elif SnowflakeGCSUtil.is_token_expired(errh.response):
                        meta.last_error = errh
                        meta.result_status = ResultStatus.RENEW_TOKEN
                    else:
                        meta.last_error = errh
                        meta.result_status = ResultStatus.ERROR
                        raise errh

        return FileHeader(
            digest=None,
            content_length=None,
            encryption_metadata=None,
        )

    @staticmethod
    def extract_bucket_name_and_path(stage_location: str) -> GcsLocation:
        container_name = stage_location
        path = ""

        # split stage location as bucket name and path
        if "/" in stage_location:
            container_name = stage_location[0 : stage_location.index("/")]
            path = stage_location[stage_location.index("/") + 1 :]
            if path and not path.endswith("/"):
                path += "/"

        return GcsLocation(bucket_name=container_name, path=path)

    @staticmethod
    def generate_file_url(stage_location: str, filename: str) -> str:
        gcs_location = SnowflakeGCSUtil.extract_bucket_name_and_path(stage_location)
        full_file_path = f"{gcs_location.path}{filename}"
        return f"https://storage.googleapis.com/{gcs_location.bucket_name}/{quote(full_file_path)}"

    @staticmethod
    def is_token_expired(response: Any) -> bool:
        # Looking further as java gcs client code, I find that token only need refresh if error is 401
        return response.status_code == 401



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2021 Snowflake Computing Inc. All right reserved.
#

from __future__ import division

import os
from logging import getLogger
from typing import TYPE_CHECKING, Any, Dict

from .constants import DEFAULT_S3_CONNECTION_POOL_SIZE, ResultStatus

if TYPE_CHECKING:  # pragma: no cover
    from .file_transfer_agent_sdk import SnowflakeFileMeta


class SnowflakeLocalUtil(object):
    @staticmethod
    def create_client(
        stage_info: Dict[str, Any],
        use_accelerate_endpoint: bool = False,
        use_s3_regional_url: bool = False,
        s3_connection_pool_size: int = DEFAULT_S3_CONNECTION_POOL_SIZE,
    ):
        return None

    @staticmethod
    def upload_one_file_with_retry(meta: "SnowflakeFileMeta") -> None:
        logger = getLogger(__name__)
        logger.debug(
            f"src_file_name=[{meta.src_file_name}], real_src_file_name=[{meta.real_src_file_name}], "
            f"stage_info=[{meta.client_meta.stage_info}], dst_file_name=[{meta.dst_file_name}]"
        )
        frd = None
        if meta.src_stream is None:
            frd = open(meta.real_src_file_name, "rb")
        else:
            frd = meta.real_src_stream or meta.src_stream
        with open(
            os.path.join(
                os.path.expanduser(meta.client_meta.stage_info["location"]),
                meta.dst_file_name,
            ),
            "wb",
        ) as output:
            output.writelines(frd)

        meta.dst_file_size = meta.upload_size
        meta.result_status = ResultStatus.UPLOADED

    @staticmethod
    def download_one_file(meta: "SnowflakeFileMeta") -> None:
        full_src_file_name = os.path.join(
            os.path.expanduser(meta.client_meta.stage_info["location"]),
            meta.src_file_name
            if not meta.src_file_name.startswith(os.sep)
            else meta.src_file_name[1:],
        )
        full_dst_file_name = os.path.join(
            meta.local_location, os.path.basename(meta.dst_file_name)
        )
        base_dir = os.path.dirname(full_dst_file_name)
        if not os.path.exists(base_dir):
            os.makedirs(base_dir)

        with open(full_src_file_name, "rb") as frd:
            with open(full_dst_file_name, "wb+") as output:
                output.writelines(frd)
        statinfo = os.stat(full_dst_file_name)
        meta.dst_file_size = statinfo.st_size
        meta.result_status = ResultStatus.DOWNLOADED



#
# Copyright (c) 2012-2021 Snowflake Computing Inc. All right reserved.
#

from __future__ import division

import json
import os
from collections import namedtuple
from logging import getLogger
from typing import TYPE_CHECKING, Any, Dict

from azure.core.exceptions import HttpResponseError, ResourceNotFoundError
from azure.storage.blob import BlobServiceClient, ContentSettings, ExponentialRetry

from .constants import HTTP_HEADER_VALUE_OCTET_STREAM, FileHeader, ResultStatus
from .encryption_util import EncryptionMetadata

if TYPE_CHECKING:  # pragma: no cover
    from .file_transfer_agent_sdk import SnowflakeFileMeta

logger = getLogger(__name__)

"""
Azure Location: Azure container name + path
"""
AzureLocation = namedtuple(
    "AzureLocation",
    ["container_name", "path"],  # Azure container name  # Azure path name
)


class SnowflakeAzureUtil(object):
    """Azure Utility class."""

    @staticmethod
    def create_client(
        stage_info: Dict[str, Any],
        use_accelerate_endpoint: bool = False,
        use_s3_regional_url: bool = False,
        s3_connection_pool_size: int = 1,
    ) -> BlobServiceClient:
        """Creates a client object with a stage credential.

        Args:
            stage_info: Information about the stage.
            use_accelerate_endpoint: Not used for Azure client.

        Returns:
            The client to communicate with GCS.
        """
        stage_credentials = stage_info["creds"]
        sas_token = stage_credentials["AZURE_SAS_TOKEN"]
        if sas_token and sas_token.startswith("?"):
            sas_token = sas_token[1:]
        end_point = stage_info["endPoint"]
        if end_point.startswith("blob."):
            end_point = end_point[len("blob.") :]
        client = BlobServiceClient(
            account_url=f"https://{stage_info['storageAccount']}.blob.{end_point}",
            credential=sas_token,
        )
        client._config.retry_policy = ExponentialRetry(
            initial_backoff=1, increment_base=2, max_attempts=60, random_jitter_range=2
        )

        return client

    @staticmethod
    def extract_container_name_and_path(stage_location):
        stage_location = os.path.expanduser(stage_location)
        container_name = stage_location
        path = ""

        # split stage location as bucket name and path
        if "/" in stage_location:
            container_name = stage_location[0 : stage_location.index("/")]
            path = stage_location[stage_location.index("/") + 1 :]
            if path and not path.endswith("/"):
                path += "/"

        return AzureLocation(container_name=container_name, path=path)

    @staticmethod
    def get_file_header(meta: "SnowflakeFileMeta", filename):
        """Gets Azure file properties."""
        client: BlobServiceClient = meta.client_meta.cloud_client
        azure_location = SnowflakeAzureUtil.extract_container_name_and_path(
            meta.client_meta.stage_info["location"]
        )
        try:
            # HTTP HEAD request
            blob = client.get_blob_client(
                azure_location.container_name, azure_location.path + filename
            )
            blob_details = blob.get_blob_properties()
        except ResourceNotFoundError:
            meta.result_status = ResultStatus.NOT_FOUND_FILE
            return FileHeader(
                digest=None, content_length=None, encryption_metadata=None
            )
        except HttpResponseError as err:
            logger.debug(
                f"Caught exception's status code: {err.status_code} and message: {str(err)}"
            )
            if (
                err.status_code == 403
                and SnowflakeAzureUtil._detect_azure_token_expire_error(err)
            ):
                logger.debug("AZURE Token expired. Renew and retry")
                meta.result_status = ResultStatus.RENEW_TOKEN
            else:
                logger.debug(
                    f"Unexpected Azure error: {err} "
                    f"container: {azure_location.container_name}, path: {azure_location.path}"
                )
                meta.result_status = ResultStatus.ERROR

            return
        meta.result_status = ResultStatus.UPLOADED
        encryptiondata = json.loads(blob_details.metadata.get("encryptiondata", "null"))
        encryption_metadata = (
            EncryptionMetadata(
                key=encryptiondata["WrappedContentKey"]["EncryptedKey"],
                iv=encryptiondata["ContentEncryptionIV"],
                matdesc=blob_details.metadata["matdesc"],
            )
            if encryptiondata
            else None
        )

        return FileHeader(
            digest=blob_details.metadata.get("sfcdigest"),
            content_length=blob_details.size,
            encryption_metadata=encryption_metadata,
        )

    @staticmethod
    def _detect_azure_token_expire_error(err):
        if err.status_code != 403:
            return False
        errstr = str(err)
        return (
            "Signature not valid in the specified time frame" in errstr
            or "Server failed to authenticate the request." in errstr
        )

    @staticmethod
    def upload_file(
        data_file: str,
        meta: "SnowflakeFileMeta",
        encryption_metadata: "EncryptionMetadata",
        max_concurrency: int,
        multipart_threshold: int,
    ):
        """Uploads the local file to Azure's Blob Storage.

        Args:
            data_file: File path on local system.
            meta: The File meta object (contains credentials and remote location).
            encryption_metadata: Encryption metadata to be set on object.
            max_concurrency: Not applicable to Azure.
            multipart_threshold: The number of bytes after which size a file should be uploaded concurrently in chunks.
                Not applicable to Azure.

        Raises:
            HTTPError if some http errors occurred.

        Returns:
            None.
        """
        azure_metadata = {
            "sfcdigest": meta.sha256_digest,
        }
        if encryption_metadata:
            azure_metadata.update(
                {
                    "encryptiondata": json.dumps(
                        {
                            "EncryptionMode": "FullBlob",
                            "WrappedContentKey": {
                                "KeyId": "symmKey1",
                                "EncryptedKey": encryption_metadata.key,
                                "Algorithm": "AES_CBC_256",
                            },
                            "EncryptionAgent": {
                                "Protocol": "1.0",
                                "EncryptionAlgorithm": "AES_CBC_128",
                            },
                            "ContentEncryptionIV": encryption_metadata.iv,
                            "KeyWrappingMetadata": {"EncryptionLibrary": "Java 5.3.0"},
                        }
                    ),
                    "matdesc": encryption_metadata.matdesc,
                }
            )
        azure_location = SnowflakeAzureUtil.extract_container_name_and_path(
            meta.client_meta.stage_info["location"]
        )
        path = azure_location.path + meta.dst_file_name.lstrip("/")

        client: BlobServiceClient = meta.client_meta.cloud_client
        callback = None
        upload_src = None
        upload_size = None

        if meta.src_stream is None:
            upload_size = os.path.getsize(data_file)
            upload_src = open(data_file, "rb")
        else:
            upload_src = meta.real_src_stream or meta.src_stream
            upload_size = upload_src.seek(0, os.SEEK_END)
            upload_src.seek(0)

        if meta.put_azure_callback:
            callback = meta.put_azure_callback(
                data_file,
                upload_size,
                output_stream=meta.put_callback_output_stream,
                show_progress_bar=meta.show_progress_bar,
            )

        def azure_callback(response):
            current = response.context["upload_stream_current"]
            total = response.context["data_stream_total"]
            if current is not None:
                callback(current)
                logger.debug(
                    "data transfer progress from sdk callback. "
                    f"current: {current}, total: {total}"
                )

        try:
            blob = client.get_blob_client(azure_location.container_name, path)
            blob.upload_blob(
                upload_src,
                metadata=azure_metadata,
                overwrite=True,
                max_concurrency=max_concurrency,
                raw_response_hook=azure_callback if meta.put_azure_callback else None,
                content_settings=ContentSettings(
                    content_type=HTTP_HEADER_VALUE_OCTET_STREAM,
                    content_encoding="utf-8",
                ),
            )
        except HttpResponseError as err:
            logger.debug(
                f"Caught exception's status code: {err.status_code} and message: {err}"
            )
            if (
                err.status_code == 403
                and SnowflakeAzureUtil._detect_azure_token_expire_error(err)
            ):
                logger.debug("AZURE Token expired. Renew and retry")
                meta.result_status = ResultStatus.RENEW_TOKEN
            else:
                meta.last_error = err
                meta.result_status = ResultStatus.NEED_RETRY
            return
        finally:
            if meta.src_stream is None:
                upload_src.close()

        logger.debug("DONE putting a file")
        meta.dst_file_size = meta.upload_size
        meta.result_status = ResultStatus.UPLOADED
        # Comparing with s3, azure haven't experienced OpenSSL.SSL.SysCallError,
        # so we will add logic to catch it only when it happens

    @staticmethod
    def _native_download_file(
        meta: "SnowflakeFileMeta", full_dst_file_name, max_concurrency
    ):
        azure_location = SnowflakeAzureUtil.extract_container_name_and_path(
            meta.client_meta.stage_info["location"]
        )
        path = azure_location.path + meta.src_file_name.lstrip("/")
        client: BlobServiceClient = meta.client_meta.cloud_client

        callback = None
        if meta.get_azure_callback:
            callback = meta.get_azure_callback(
                meta.src_file_name,
                meta.src_file_size,
                output_stream=meta.get_callback_output_stream,
                show_progress_bar=meta.show_progress_bar,
            )

        def azure_callback(response):
            current = response.context["download_stream_current"]
            total = response.context["data_stream_total"]
            if current is not None:
                callback(current)
                logger.debug(
                    f"data transfer progress from sdk callback. current: {current}, total: {total}"
                )

        try:
            blob = client.get_blob_client(azure_location.container_name, path)
            with open(full_dst_file_name, "wb") as download_f:
                download = blob.download_blob(
                    max_concurrency=max_concurrency,
                    raw_response_hook=azure_callback
                    if meta.put_azure_callback
                    else None,
                )
                download.readinto(download_f)

        except HttpResponseError as err:
            logger.debug(
                f"Caught exception's status code: {err.status_code} and message: {str(err)}"
            )
            if (
                err.status_code == 403
                and SnowflakeAzureUtil._detect_azure_token_expire_error(err)
            ):
                logger.debug("AZURE Token expired. Renew and retry")
                meta.result_status = ResultStatus.RENEW_TOKEN
            else:
                meta.last_error = err
                meta.result_status = ResultStatus.NEED_RETRY
            return
        meta.result_status = ResultStatus.DOWNLOADED



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2021 Snowflake Computing Inc. All right reserved.
#

from datetime import datetime, time, timedelta
from logging import getLogger

import pytz

from .converter import ZERO_EPOCH, SnowflakeConverter, _generate_tzinfo_from_tzoffset

logger = getLogger(__name__)


class SnowflakeConverterIssue23517(SnowflakeConverter):
    """Converter for Python 3.5.0 or Any Python on Windows.

    This is to address http://bugs.python.org/issue23517
    """

    def __init__(self, **kwargs):
        super(SnowflakeConverterIssue23517, self).__init__(**kwargs)
        logger.debug("initialized")

    def _TIMESTAMP_TZ_to_python(self, ctx):
        """Converts TIMESTAMP TZ to datetime.

        The timezone offset is piggybacked.
        """
        scale = ctx["scale"]

        def conv0(encoded_value: str) -> datetime:
            value, tz = encoded_value.split()
            tzinfo = _generate_tzinfo_from_tzoffset(int(tz) - 1440)
            microseconds = float(value)
            t = ZERO_EPOCH + timedelta(seconds=microseconds)
            if pytz.utc != tzinfo:
                t += tzinfo.utcoffset(t)
            return t.replace(tzinfo=tzinfo)

        def conv(encoded_value: str) -> datetime:
            value, tz = encoded_value.split()
            tzinfo = _generate_tzinfo_from_tzoffset(int(tz) - 1440)
            microseconds = float(value[0 : -scale + 6])
            t = ZERO_EPOCH + timedelta(seconds=microseconds)
            if pytz.utc != tzinfo:
                t += tzinfo.utcoffset(t)
            return t.replace(tzinfo=tzinfo)

        return conv if scale > 6 else conv0

    def _TIMESTAMP_NTZ_to_python(self, ctx):
        """Converts TIMESTAMP NTZ to datetime.

        No timezone info is attached.
        """
        scale = ctx["scale"]

        def conv0(value: str) -> datetime:
            logger.debug("timestamp_ntz: %s", value)
            return ZERO_EPOCH + timedelta(seconds=(float(value)))

        def conv(value: str) -> datetime:
            logger.debug("timestamp_ntz: %s", value)
            microseconds = float(value[0 : -scale + 6])
            return ZERO_EPOCH + timedelta(seconds=(microseconds))

        return conv if scale > 6 else conv0

    def _TIMESTAMP_LTZ_to_python(self, ctx):
        def conv(value: str) -> datetime:
            t, _ = self._pre_TIMESTAMP_LTZ_to_python(value, ctx)
            return t

        return conv

    def _TIME_to_python(self, ctx):
        """Converts TIME to formatted string, SnowflakeDateTime, or datetime.time.

        No timezone is attached.
        """
        scale = ctx["scale"]

        def conv0(value: str) -> time:
            return (ZERO_EPOCH + timedelta(seconds=(float(value)))).time()

        def conv(value: str) -> time:
            microseconds = float(value[0 : -scale + 6])
            return (ZERO_EPOCH + timedelta(seconds=(microseconds))).time()

        return conv if scale > 6 else conv0
